import copy
from metadrive.scenario.parse_object_state import parse_full_trajectory, parse_object_state
from collections import namedtuple, OrderedDict

import numpy as np
from metadrive.type import MetaDriveType
from metadrive.component.lane.point_lane import PointLane
from metadrive.component.vehicle.vehicle_type import SVehicle
from metadrive.manager.waymo_traffic_manager import WaymoTrafficManager
from metadrive.policy.idm_policy import WaymoIDMPolicy

static_vehicle_info = namedtuple("static_vehicle_info", "position heading")


def handler(signum, frame):
    raise Exception("end of time")


class WaymoIDMTrafficManager(WaymoTrafficManager):
    TRAJ_WIDTH = 1.2
    DEST_REGION = 5
    MIN_DURATION = 20
    ACT_FREQ = 5
    MAX_HORIZON = 100

    def __init__(self):
        raise DeprecationWarning(
            "This class is deprecated now, "
            "it is merged into WaymoTrafficManager (ScenarioTrafficManager)"
        )
        super(WaymoIDMTrafficManager, self).__init__()
        self.seed_trajs = {}
        self.v_id_to_destination = OrderedDict()
        self.v_id_to_stop_time = OrderedDict()

    def before_reset(self):
        super(WaymoIDMTrafficManager, self).before_reset()

    def after_reset(self):
        # try:
        self.v_id_to_destination = OrderedDict()
        self.v_id_to_stop_time = OrderedDict()
        self.count = 0
        if self.engine.global_random_seed not in self.seed_trajs:
            traffic_traj_data = {}
            for v_id, type_traj in self.current_traffic_data.items():
                if MetaDriveType.is_vehicle(type_traj["type"]) and v_id != self.sdc_track_index:
                    init_info = parse_object_state(type_traj, self.engine.global_config["traj_start_index"])
                    dest_info = parse_object_state(
                        type_traj,
                        self.engine.global_config["traj_end_index"],
                        check_last_state=True,
                    )
                    if not init_info["valid"]:
                        continue
                    if np.linalg.norm(np.array(init_info["position"]) - np.array(dest_info["position"])) < 1:
                        full_traj = static_vehicle_info(init_info["position"], init_info["heading"])
                        static = True
                    else:
                        full_traj = parse_full_trajectory(type_traj)
                        if len(full_traj) < self.MIN_DURATION:
                            full_traj = static_vehicle_info(init_info["position"], init_info["heading"])
                            static = True
                        else:
                            full_traj = PointLane(full_traj, width=self.TRAJ_WIDTH)
                            static = False
                    traffic_traj_data[v_id] = {
                        "traj": full_traj,
                        "init_info": init_info,
                        "static": static,
                        "dest_info": dest_info,
                        "is_sdc": False
                    }

                elif MetaDriveType.is_vehicle(type_traj["type"]) and v_id == self.sdc_track_index:
                    # set Ego V velocity
                    init_info = parse_object_state(
                        type_traj,
                        self.engine.global_config["traj_start_index"],
                    )
                    traffic_traj_data["sdc"] = {
                        "traj": None,
                        "init_info": init_info,
                        "static": False,
                        "dest_info": None,
                        "is_sdc": True
                    }
            self.seed_trajs[self.engine.global_random_seed] = traffic_traj_data
        policy_count = 0
        for v_traj_id, data in self.current_traffic_traj.items():
            if data["static"] and self.engine.global_config["no_static_vehicles"]:
                continue
            if v_traj_id == "sdc":
                continue
            init_info = data["init_info"]
            v_config = copy.deepcopy(self.engine.global_config["vehicle_config"])
            v_config.update(
                dict(
                    show_navi_mark=False,
                    show_dest_mark=False,
                    enable_reverse=False,
                    show_lidar=False,
                    show_lane_line_detector=False,
                    show_side_detector=False,
                    need_navigation=False,
                )
            )
            v = self.spawn_object(
                SVehicle, position=init_info["position"], heading=init_info["heading"], vehicle_config=v_config
            )
            self.v_id_to_destination[v.id] = np.array(data["dest_info"]["position"])
            self.v_id_to_stop_time[v.id] = 0
            if data["static"]:
                # static vehicle
                v.set_position(v.position, height=0.8)
                v.set_velocity((0, 0))
                v.set_static(True)
            else:
                v.set_position(v.position, height=0.8)
                self.add_policy(
                    v.id, WaymoIDMPolicy, v, self.generate_seed(), data["traj"], policy_count % self.ACT_FREQ
                )
                v.set_velocity(init_info['velocity'])
                policy_count += 1

    def before_step(self, *args, **kwargs):
        for v in self.spawned_objects.values():
            if self.engine.has_policy(v.id):
                p = self.engine.get_policy(v.name)
                do_speed_control = (p.policy_index + self.count) % self.ACT_FREQ == 0
                v.before_step(p.act(do_speed_control))

    def after_step(self, *args, **kwargs):
        self.count += 1
        vehicles_to_clear = []
        # LQY: modify termination condition
        for v in self.spawned_objects.values():
            if not self.engine.has_policy(v.name):
                continue
            if v.speed_km_h < 1:
                self.v_id_to_stop_time[v.id] += 1
            else:
                self.v_id_to_stop_time[v.id] = 0
            dist_to_dest = np.linalg.norm(v.position - self.v_id_to_destination[v.id])
            if dist_to_dest < self.DEST_REGION or self.v_id_to_stop_time[v.id] > self.MAX_HORIZON:
                vehicles_to_clear.append(v.id)
        self.clear_objects(vehicles_to_clear)

    @property
    def current_traffic_traj(self):
        return self.seed_trajs[self.engine.global_random_seed]



"""
This file illustrate how to use top-down renderer to provide observation in form of multiple channels of semantic maps.

We let the target vehicle moving forward directly. You can also try to control the vehicle by yourself. See the config
below for more information.

This script will popup a Pygame window, but that is not the form of the observation. We will also popup a matplotlib
window, which shows the details observation of the top-down pygame renderer.

The detailed implementation of the Pygame renderer is in TopDownMultiChannel Class (a subclass of Observation Class)
at: metadrive/obs/top_down_obs_multi_channel.py

We welcome contributions to propose a better representation of the top-down semantic observation!
"""

import random

import matplotlib.pyplot as plt

from metadrive import TopDownMetaDrive
from metadrive.constants import HELP_MESSAGE
from metadrive.examples.ppo_expert.numpy_expert import expert


def draw_multi_channels_top_down_observation(obs, show_time=4):
    num_channels = obs.shape[-1]
    assert num_channels == 5
    channel_names = [
        "Road and navigation", "Ego now and previous pos", "Neighbor at step t", "Neighbor at step t-1",
        "Neighbor at step t-2"
    ]
    fig, axs = plt.subplots(1, num_channels, figsize=(15, 4), dpi=80)
    count = 0

    def close_event():
        plt.close()  # timer calls this function after 3 seconds and closes the window

    timer = fig.canvas.new_timer(
        interval=show_time * 1000
    )  # creating a timer object and setting an interval of 3000 milliseconds
    timer.add_callback(close_event)

    for i, name in enumerate(channel_names):
        count += 1
        ax = axs[i]
        ax.imshow(obs[..., i], cmap="bone")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title(name)
        # print("Drawing {}-th semantic map!".format(count))
    fig.suptitle("Multi-channels Top-down Observation")
    timer.start()
    plt.show()


if __name__ == "__main__":
    print(HELP_MESSAGE)
    env = TopDownMetaDrive(
        dict(
            # We also support using two renderer (Panda3D renderer and Pygame renderer) simultaneously. You can
            # try this by uncommenting next line.
            # use_render=True,

            # You can also try to uncomment next line with "use_render=True", so that you can control the ego vehicle
            # with keyboard in the main window.
            # manual_control=True,
            map="SSSS",
            traffic_density=0.1,
            num_scenarios=100,
            start_seed=random.randint(0, 1000),
        )
    )
    try:
        o, _ = env.reset()
        for i in range(1, 100000):
            o, r, tm, tc, info = env.step(expert(env.vehicle))
            env.render(mode="top_down", text={"Quit": "ESC"}, film_size=(2000, 2000))
            if tm or tc:
                env.reset()
            if i % 50 == 0:
                draw_multi_channels_top_down_observation(o, show_time=5)  # show time 4s
                # ret = input("Do you wish to quit? Type any ESC to quite, or press enter to continue")
                # if len(ret) == 0:
                #     continue
                # else:
                #     break
    except Exception as e:
        raise e
    finally:
        env.close()



import re
from typing import List, Tuple

from metadrive.constants import Decoration

LaneIndex = Tuple[str, str, int]
Route = List[LaneIndex]


class Road:
    """
    Road is a bunch of lanes connecting two nodes, one start and the other end
    """
    NEGATIVE_DIR = "-"

    def __init__(self, start_node: str, end_node: str):
        self.start_node = start_node
        self.end_node = end_node

    def get_lanes(self, road_network):
        return road_network.graph[self.start_node][self.end_node]

    def __neg__(self):
        sub_index = self.end_node.find(Road.NEGATIVE_DIR)
        if sub_index == -1:
            return Road(Road.NEGATIVE_DIR + self.end_node, Road.NEGATIVE_DIR + self.start_node)
        else:
            return Road(self.end_node[sub_index + 1:], self.start_node[sub_index + 1:])

    def is_negative_road(self):
        return False if self.end_node.find(Road.NEGATIVE_DIR) == -1 else True

    def is_valid_road(self):
        return False if self.start_node == Decoration.start and self.end_node == Decoration.end else True

    def lane_index(self, index: int) -> LaneIndex:
        return self.start_node, self.end_node, index

    def lane_num(self, road_network):
        return len(self.get_lanes(road_network))

    def block_ID(self):
        search_node = self.end_node if not self.is_negative_road() else self.start_node
        if re.search(">", search_node) is not None:
            return ">"
        block_id = re.search("[a-zA-Z$]", search_node).group(0)
        return block_id

    def __eq__(self, other):
        if isinstance(other, Road):
            return True if self.start_node == other.start_node and self.end_node == other.end_node else False
        else:
            return super(Road, self).__eq__(other)

    def __repr__(self):
        return "Road from {} to {}".format(self.start_node, self.end_node)

    def __hash__(self):
        return hash((self.start_node, self.end_node))

    def to_json(self):
        return (self.start_node, self.end_node)



import numpy as np

from metadrive.component.lane.circular_lane import CircularLane
from metadrive.component.pg_space import Parameter, BlockParameterSpace
from metadrive.component.road_network.edge_road_network import EdgeRoadNetwork
from metadrive.component.vehicle_navigation_module.base_navigation import BaseNavigation
from metadrive.utils import clip, norm
from metadrive.utils.math import panda_vector
from metadrive.utils.pg.utils import ray_localization


class EdgeNetworkNavigation(BaseNavigation):
    """
   This class define a helper for localizing vehicles and retrieving navigation information.
   It now only support EdgeRoadNetwork
   """
    def __init__(
        self,
        show_navi_mark: bool = False,
        random_navi_mark_color=False,
        show_dest_mark=False,
        show_line_to_dest=False,
        panda_color=None,
        name=None,
        vehicle_config=None
    ):
        super(EdgeNetworkNavigation, self).__init__(
            show_navi_mark=show_navi_mark,
            random_navi_mark_color=random_navi_mark_color,
            show_dest_mark=show_dest_mark,
            show_line_to_dest=show_line_to_dest,
            panda_color=panda_color,
            name=name,
            vehicle_config=vehicle_config
        )

    def reset(self, vehicle):
        possible_lanes = ray_localization(vehicle.heading, vehicle.spawn_place, self.engine, use_heading_filter=False)
        possible_lane_indexes = [lane_index for lane, lane_index, dist in possible_lanes]

        if len(possible_lanes) == 0 and vehicle.config["spawn_lane_index"] is None:
            from metadrive.utils.error_class import NavigationError
            raise NavigationError("Can't find valid lane for navigation.")

        if vehicle.config["spawn_lane_index"] is not None and vehicle.config["spawn_lane_index"
                                                                             ] in possible_lane_indexes:
            idx = possible_lane_indexes.index(vehicle.config["spawn_lane_index"])
            lane, new_l_index = possible_lanes[idx][:-1]
        else:
            assert len(possible_lanes) > 0
            lane, new_l_index = possible_lanes[0][:-1]

        dest = vehicle.config["destination"]

        current_lane = lane
        destination = dest if dest is not None else None
        assert current_lane is not None, "spawn place is not on road!"

        super(EdgeNetworkNavigation, self).reset(current_lane)
        assert self.map.road_network_type == EdgeRoadNetwork, "This Navigation module only support EdgeRoadNetwork type"
        self.set_route(current_lane.index, destination)

    def set_route(self, current_lane_index: str, destination: str):
        """
        Find a shortest path from start road to end road
        :param current_lane_index: start road node
        :param destination: end road node or end lane index
        :return: None
        """
        self.checkpoints = self.map.road_network.shortest_path(current_lane_index, destination)
        # if len(self.checkpoints) == 0:
        #     self.checkpoints.append(current_lane_index)
        #     self.checkpoints.append(current_lane_index)
        self._target_checkpoints_index = [0, 1]
        # update routing info
        assert len(self.checkpoints) > 0, "Can not find a route from {} to {}".format(current_lane_index, destination)
        self.final_lane = self.map.road_network.get_lane(self.checkpoints[-1])
        self._navi_info.fill(0.0)
        self.current_ref_lanes = self.map.road_network.get_peer_lanes_from_index(self.current_checkpoint_lane_index)
        self.next_ref_lanes = self.map.road_network.get_peer_lanes_from_index(self.next_checkpoint_lane_index)
        if self._dest_node_path is not None:
            ref_lane = self.final_lane
            later_middle = (float(self.get_current_lane_num()) / 2 - 0.5) * self.get_current_lane_width()
            check_point = ref_lane.position(ref_lane.length, later_middle)
            # self._dest_node_path.setPos(check_point[0], -check_point[1], 1.8)
            self._dest_node_path.setPos(panda_vector(check_point[0], check_point[1], 1.8))

    def update_localization(self, ego_vehicle):
        position = ego_vehicle.position
        lane, lane_index = self._update_current_lane(ego_vehicle)
        need_update = self._update_target_checkpoints(lane_index)

        # target_road_1 is the road segment the vehicle is driving on.
        if need_update:
            self.current_ref_lanes = self.map.road_network.get_peer_lanes_from_index(self.current_checkpoint_lane_index)

            # target_road_2 is next road segment the vehicle should drive on.
            self.next_ref_lanes = self.map.road_network.get_peer_lanes_from_index(self.next_checkpoint_lane_index)

            if self.current_checkpoint_lane_index == self.next_checkpoint_lane_index:
                # When we are in the final road segment that there is no further road to drive on
                self.next_ref_lanes = None

        self._navi_info.fill(0.0)
        half = self.CHECK_POINT_INFO_DIM
        self._navi_info[:half], lanes_heading1, checkpoint = self._get_info_for_checkpoint(
            lanes_id=0,
            ref_lane=self.map.road_network.get_lane(self.current_checkpoint_lane_index),
            ego_vehicle=ego_vehicle
        )

        self._navi_info[half:], lanes_heading2, _ = self._get_info_for_checkpoint(
            lanes_id=1,
            ref_lane=self.map.road_network.get_lane(self.next_checkpoint_lane_index),
            ego_vehicle=ego_vehicle
        )

        if self._show_navi_info:  # Whether to visualize little boxes in the scene denoting the checkpoints
            pos_of_goal = checkpoint
            self._goal_node_path.setPos(panda_vector(pos_of_goal[0], pos_of_goal[1], 1.8))
            self._goal_node_path.setH(self._goal_node_path.getH() + 3)
            self.navi_arrow_dir = [lanes_heading1, lanes_heading2]
            dest_pos = self._dest_node_path.getPos()
            self._draw_line_to_dest(start_position=ego_vehicle.position, end_position=(dest_pos[0], dest_pos[1]))
            navi_pos = self._goal_node_path.getPos()
            self._draw_line_to_navi(start_position=ego_vehicle.position, end_position=(navi_pos[0], navi_pos[1]))

    def _update_target_checkpoints(self, ego_lane_index) -> bool:
        """
        update the checkpoint, return True if updated else False
        """
        if self.current_checkpoint_lane_index == self.next_checkpoint_lane_index:  # on last road
            return False

        # arrive to second checkpoint
        new_index = ego_lane_index
        if new_index in self.checkpoints[self._target_checkpoints_index[1]:]:
            idx = self.checkpoints.index(new_index, self._target_checkpoints_index[1])
            self._target_checkpoints_index = [idx]
            if idx + 1 == len(self.checkpoints):
                self._target_checkpoints_index.append(idx)
            else:
                self._target_checkpoints_index.append(idx + 1)
            return True
        return False

    def get_current_lateral_range(self, current_position, engine) -> float:
        return self.current_lane.width * len(self.current_ref_lanes)

    @property
    def current_checkpoint_lane_index(self):
        return self.checkpoints[self._target_checkpoints_index[0]]

    @property
    def next_checkpoint_lane_index(self):
        return self.checkpoints[self._target_checkpoints_index[1]]

    def _get_current_lane(self, ego_vehicle):
        """
        Called in update_localization to find current lane information
        """
        possible_lanes, on_lane = ray_localization(
            ego_vehicle.heading,
            ego_vehicle.position,
            ego_vehicle.engine,
            use_heading_filter=False,
            return_on_lane=True
        )
        for lane, index, l_1_dist in possible_lanes:
            if lane in self.current_ref_lanes:
                return lane, index, on_lane
        nx_ckpt = self._target_checkpoints_index[-1]
        if nx_ckpt == self.checkpoints[-1] or self.next_ref_lanes is None:
            return (*possible_lanes[0][:-1], on_lane) if len(possible_lanes) > 0 else (None, None, on_lane)

        next_ref_lanes = self.next_ref_lanes
        for lane, index, l_1_dist in possible_lanes:
            if lane in next_ref_lanes:
                return lane, index, on_lane
        return (*possible_lanes[0][:-1], on_lane) if len(possible_lanes) > 0 else (None, None, on_lane)

    def _get_info_for_checkpoint(self, lanes_id, ref_lane, ego_vehicle):

        navi_information = []
        # Project the checkpoint position into the target vehicle's coordination, where
        # +x is the heading and +y is the right hand side.
        check_point = ref_lane.end
        dir_vec = check_point - ego_vehicle.position  # get the vector from center of vehicle to checkpoint
        dir_norm = norm(dir_vec[0], dir_vec[1])
        if dir_norm > self.NAVI_POINT_DIST:  # if the checkpoint is too far then crop the direction vector
            dir_vec = dir_vec / dir_norm * self.NAVI_POINT_DIST
        ckpt_in_heading, ckpt_in_rhs = ego_vehicle.convert_to_local_coordinates(
            dir_vec, 0.0
        )  # project to ego vehicle's coordination

        # Dim 1: the relative position of the checkpoint in the target vehicle's heading direction.
        navi_information.append(clip((ckpt_in_heading / self.NAVI_POINT_DIST + 1) / 2, 0.0, 1.0))

        # Dim 2: the relative position of the checkpoint in the target vehicle's right hand side direction.
        navi_information.append(clip((ckpt_in_rhs / self.NAVI_POINT_DIST + 1) / 2, 0.0, 1.0))

        if lanes_id == 0:
            lanes_heading = ref_lane.heading_theta_at(ref_lane.local_coordinates(ego_vehicle.position)[0])
        else:
            lanes_heading = ref_lane.heading_theta_at(min(self.PRE_NOTIFY_DIST, ref_lane.length))

        # Try to include the current lane's information into the navigation information
        bendradius = 0.0
        dir = 0.0
        angle = 0.0
        if isinstance(ref_lane, CircularLane):
            bendradius = ref_lane.radius / (
                BlockParameterSpace.CURVE[Parameter.radius].max +
                self.get_current_lane_num() * self.get_current_lane_width()
            )
            dir = -ref_lane.direction
            angle = ref_lane.angle
        # Dim 3: The bending radius of current lane
        navi_information.append(clip(bendradius, 0.0, 1.0))

        # Dim 4: The bending direction of current lane (+1 for clockwise, -1 for counterclockwise)
        navi_information.append(clip((dir + 1) / 2, 0.0, 1.0))

        # Dim 5: The angular difference between the heading in lane ending position and
        # the heading in lane starting position
        navi_information.append(
            clip((np.rad2deg(angle) / BlockParameterSpace.CURVE[Parameter.angle].max + 1) / 2, 0.0, 1.0)
        )
        return navi_information, lanes_heading, check_point

    def _update_current_lane(self, ego_vehicle):
        lane, lane_index, on_lane = self._get_current_lane(ego_vehicle)
        ego_vehicle.on_lane = on_lane
        if lane is None:
            lane, lane_index = ego_vehicle.lane, ego_vehicle.lane_index
            if self.FORCE_CALCULATE:
                lane_index, _ = self.map.road_network.get_closest_lane_index(ego_vehicle.position)
                lane = self.map.road_network.get_lane(lane_index)
        self.current_lane = lane
        assert lane_index == lane.index, "lane index mismatch!"
        return lane, lane_index



ALL_TYPE = {
    "noise": 'noise',
    "human.pedestrian.adult": 'adult',
    "human.pedestrian.child": 'child',
    "human.pedestrian.wheelchair": 'wheelchair',
    "human.pedestrian.stroller": 'stroller',
    "human.pedestrian.personal_mobility": 'p.mobility',
    "human.pedestrian.police_officer": 'police',
    "human.pedestrian.construction_worker": 'worker',
    "animal": 'animal',
    "vehicle.car": 'car',
    "vehicle.motorcycle": 'motorcycle',
    "vehicle.bicycle": 'bicycle',
    "vehicle.bus.bendy": 'bus.bendy',
    "vehicle.bus.rigid": 'bus.rigid',
    "vehicle.truck": 'truck',
    "vehicle.construction": 'constr. veh',
    "vehicle.emergency.ambulance": 'ambulance',
    "vehicle.emergency.police": 'police car',
    "vehicle.trailer": 'trailer',
    "movable_object.barrier": 'barrier',
    "movable_object.trafficcone": 'trafficcone',
    "movable_object.pushable_pullable": 'push/pullable',
    "movable_object.debris": 'debris',
    "static_object.bicycle_rack": 'bicycle racks',
    "flat.driveable_surface": 'driveable',
    "flat.sidewalk": 'sidewalk',
    "flat.terrain": 'terrain',
    "flat.other": 'flat.other',
    "static.manmade": 'manmade',
    "static.vegetation": 'vegetation',
    "static.other": 'static.other',
    "vehicle.ego": "ego"
}
NOISE_TYPE = {
    "noise": 'noise',
    "animal": 'animal',
    "static_object.bicycle_rack": 'bicycle racks',
    "movable_object.pushable_pullable": 'push/pullable',
    "movable_object.debris": 'debris',
    "static.manmade": 'manmade',
    "static.vegetation": 'vegetation',
    "static.other": 'static.other',
}
HUMAN_TYPE = {
    "human.pedestrian.adult": 'adult',
    "human.pedestrian.child": 'child',
    "human.pedestrian.wheelchair": 'wheelchair',
    "human.pedestrian.stroller": 'stroller',
    "human.pedestrian.personal_mobility": 'p.mobility',
    "human.pedestrian.police_officer": 'police',
    "human.pedestrian.construction_worker": 'worker',
}
BICYCLE_TYPE = {
    "vehicle.bicycle": 'bicycle',
    "vehicle.motorcycle": 'motorcycle',
}
VEHICLE_TYPE = {
    "vehicle.car": 'car',
    "vehicle.bus.bendy": 'bus.bendy',
    "vehicle.bus.rigid": 'bus.rigid',
    "vehicle.truck": 'truck',
    "vehicle.construction": 'constr. veh',
    "vehicle.emergency.ambulance": 'ambulance',
    "vehicle.emergency.police": 'police car',
    "vehicle.trailer": 'trailer',
    "vehicle.ego": "ego",
}
OBSTACLE_TYPE = {
    "movable_object.barrier": 'barrier',
    "movable_object.trafficcone": 'trafficcone',
}
TERRAIN_TYPE = {
    "flat.driveable_surface": 'driveable',
    "flat.sidewalk": 'sidewalk',
    "flat.terrain": 'terrain',
    "flat.other": 'flat.other'
}



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

setup_logger(debug=True)


def _test_destroy(config):
    env = MetaDriveEnv(config)
    try:
        env.reset()
        for i in range(1, 20):
            env.step([1, 1])

        env.close()
        env.reset()
        env.close()
        env.close()
        env.reset()
        env.reset()
        env.close()

        # Again!
        env = MetaDriveEnv(config)
        env.reset()
        for i in range(1, 20):
            env.step([1, 1])
        env.reset()
        env.close()
    finally:
        env.close()


def _test_destroy_rgb(obs="rgb"):
    # Close and reset
    config = {"num_scenarios": 1, "start_seed": 3, "manual_control": False}
    if obs == "state":
        pass
    elif obs == "rgb":
        config["image_observation"] = True
    else:
        config["use_render"] = True
    _test_destroy(config)


def test_destroy_state(obs="state"):
    # Close and reset
    config = {"num_scenarios": 1, "start_seed": 3, "manual_control": False}
    if obs == "state":
        pass
    elif obs == "rgb":
        config["image_observation"] = True
        config["use_render"] = False
        config["rgb_clip"] = True
    _test_destroy(config)


if __name__ == "__main__":
    # test_destroy_rgb()
    test_destroy_state()



"""
This environment can load all scenarios exported from other environments via env.export_scenarios()
"""
import logging

import numpy as np

from metadrive.component.sensors.vehicle_panel import VehiclePanel
from metadrive.component.vehicle_navigation_module.trajectory_navigation import TrajectoryNavigation
from metadrive.constants import TerminationState
from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.base_env import BaseEnv
from metadrive.manager.scenario_curriculum_manager import ScenarioCurriculumManager
from metadrive.manager.scenario_data_manager import ScenarioDataManager
from metadrive.manager.scenario_light_manager import ScenarioLightManager
from metadrive.manager.scenario_map_manager import ScenarioMapManager
from metadrive.manager.waymo_traffic_manager import WaymoTrafficManager
from metadrive.obs.image_obs import ImageStateObservation
from metadrive.obs.state_obs import LidarStateObservation
from metadrive.policy.replay_policy import ReplayEgoCarPolicy
from metadrive.utils import get_np_random
from metadrive.utils.math import wrap_to_pi

SCENARIO_ENV_CONFIG = dict(
    # ===== Scenario Config =====
    data_directory=AssetLoader.file_path("waymo", return_raw_style=False),
    start_scenario_index=0,
    num_scenarios=3,
    sequential_seed=False,  # Whether to set seed (the index of map) sequentially across episodes
    worker_index=0,  # Allowing multi-worker sampling with Rllib
    num_workers=1,  # Allowing multi-worker sampling with Rllib

    # ===== Curriculum Config =====
    curriculum_level=1,  # i.e. set to 5 to split the data into 5 difficulty level
    episodes_to_evaluate_curriculum=None,
    target_success_rate=0.8,

    # ===== Map Config =====
    store_map=True,
    store_data=True,
    need_lane_localization=True,
    no_map=False,

    # ===== Traffic =====
    no_traffic=False,  # nothing will be generated including objects/pedestrian/vehicles
    no_static_vehicles=False,  # static vehicle will be removed
    no_light=False,  # no traffic light
    reactive_traffic=False,  # turn on to enable idm traffic
    filter_overlapping_car=True,  # If in one frame a traffic vehicle collides with ego car, it won't be created.
    even_sample_vehicle_class=True,  # to make the scene more diverse
    default_vehicle_in_traffic=False,
    skip_missing_light=False,
    static_traffic_object=True,

    # ===== Agent config =====
    vehicle_config=dict(
        lidar=dict(num_lasers=120, distance=50),
        lane_line_detector=dict(num_lasers=0, distance=50),
        side_detector=dict(num_lasers=12, distance=50),
        show_dest_mark=True,
        navigation_module=TrajectoryNavigation,
    ),

    # ===== Reward Scheme =====
    # See: https://github.com/metadriverse/metadrive/issues/283
    success_reward=5.0,
    out_of_road_penalty=5.0,
    on_lane_line_penalty=1.,
    crash_vehicle_penalty=1.,
    crash_object_penalty=1.0,
    crash_human_penalty=1.0,
    driving_reward=1.0,
    steering_range_penalty=0.5,
    heading_penalty=1.0,
    lateral_penalty=.5,
    max_lateral_dist=4,
    no_negative_reward=True,

    # ===== Cost Scheme =====
    crash_vehicle_cost=1.0,
    crash_object_cost=1.0,
    out_of_road_cost=1.0,
    crash_human_cost=1.0,

    # ===== Termination Scheme =====
    out_of_route_done=False,
    crash_vehicle_done=False,
    crash_object_done=False,
    crash_human_done=False,
    relax_out_of_road_done=True,

    # ===== others =====
    interface_panel=["dashboard"],  # for boosting efficiency
    horizon=None,
    allowed_more_steps=None,  # None=infinite
)


class ScenarioEnv(BaseEnv):
    @classmethod
    def default_config(cls):
        config = super(ScenarioEnv, cls).default_config()
        config.update(SCENARIO_ENV_CONFIG)
        return config

    def __init__(self, config=None):
        super(ScenarioEnv, self).__init__(config)
        if self.config["curriculum_level"] > 1:
            assert self.config["num_scenarios"] % self.config["curriculum_level"] == 0, \
                "Each level should have the same number of scenarios"
            if self.config["num_workers"] > 1:
                num = int(self.config["num_scenarios"] / self.config["curriculum_level"])
                assert num % self.config["num_workers"] == 0
        if self.config["num_workers"] > 1:
            assert self.config["sequential_seed"], \
                "If using > 1 workers, you have to allow sequential_seed for consistency!"

    def _get_observations(self):
        return {self.DEFAULT_AGENT: self.get_single_observation()}

    def get_single_observation(self):
        if self.config["image_observation"]:
            o = ImageStateObservation(self.config)
        else:
            o = LidarStateObservation(self.config)
        return o

    def switch_to_top_down_view(self):
        self.main_camera.stop_track()

    def switch_to_third_person_view(self):
        if self.main_camera is None:
            return
        self.main_camera.reset()
        if self.config["prefer_track_agent"] is not None and self.config["prefer_track_agent"] in self.vehicles.keys():
            new_v = self.vehicles[self.config["prefer_track_agent"]]
            current_track_vehicle = new_v
        else:
            if self.main_camera.is_bird_view_camera():
                current_track_vehicle = self.current_track_vehicle
            else:
                vehicles = list(self.engine.agents.values())
                if len(vehicles) <= 1:
                    return
                if self.current_track_vehicle in vehicles:
                    vehicles.remove(self.current_track_vehicle)
                new_v = get_np_random().choice(vehicles)
                current_track_vehicle = new_v
        self.main_camera.track(current_track_vehicle)
        return

    def setup_engine(self):
        self.in_stop = False
        super(ScenarioEnv, self).setup_engine()
        self.engine.register_manager("data_manager", ScenarioDataManager())
        self.engine.register_manager("map_manager", ScenarioMapManager())
        if not self.config["no_traffic"]:
            self.engine.register_manager("traffic_manager", WaymoTrafficManager())
        if not self.config["no_light"]:
            self.engine.register_manager("light_manager", ScenarioLightManager())
        self.engine.register_manager("curriculum_manager", ScenarioCurriculumManager())
        self.engine.accept("p", self.stop)
        self.engine.accept("q", self.switch_to_third_person_view)
        self.engine.accept("b", self.switch_to_top_down_view)
        self.engine.accept("]", self.next_seed_reset)
        self.engine.accept("[", self.last_seed_reset)

    def next_seed_reset(self):
        if self.current_seed + 1 < self.config["start_scenario_index"] + self.config["num_scenarios"]:
            self.reset(self.current_seed + 1)
        else:
            logging.warning("Can't load next scenario! current seed is already the max scenario index")

    def last_seed_reset(self):
        if self.current_seed - 1 >= self.config["start_scenario_index"]:
            self.reset(self.current_seed - 1)
        else:
            logging.warning("Can't load last scenario! current seed is already the min scenario index")

    def step(self, actions):
        ret = super(ScenarioEnv, self).step(actions)
        while self.in_stop:
            self.engine.taskMgr.step()
        return ret

    def done_function(self, vehicle_id: str):
        vehicle = self.vehicles[vehicle_id]
        done = False
        done_info = dict(
            crash_vehicle=False,
            crash_object=False,
            crash_building=False,
            out_of_road=False,
            arrive_dest=False,
            max_step=False,
        )

        route_completion = vehicle.navigation.route_completion

        def msg(reason):
            return "Episode ended! Scenario Index: {} Scenario id: {} Reason: {}.".format(
                self.current_seed, self.engine.data_manager.current_scenario_id, reason
            )

        if self._is_arrive_destination(vehicle):
            done = True
            logging.info(msg("arrive_dest"))
            done_info[TerminationState.SUCCESS] = True

        elif self._is_out_of_road(vehicle) or route_completion < -0.1:
            done = True
            logging.info(msg("out_of_road"))
            done_info[TerminationState.OUT_OF_ROAD] = True
        elif vehicle.crash_human and self.config["crash_human_done"]:
            done = True
            logging.info(msg("crash human"))
            done_info[TerminationState.CRASH_HUMAN] = True
        elif vehicle.crash_vehicle and self.config["crash_vehicle_done"]:
            done = True
            logging.info(msg("crash vehicle"))
            done_info[TerminationState.CRASH_VEHICLE] = True
        elif vehicle.crash_object and self.config["crash_object_done"]:
            done = True
            done_info[TerminationState.CRASH_OBJECT] = True
            logging.info(msg("crash object"))
        elif vehicle.crash_building and self.config["crash_object_done"]:
            done = True
            done_info[TerminationState.CRASH_BUILDING] = True
            logging.info(msg("crash building"))

        elif self.config["horizon"] is not None and \
                self.episode_lengths[vehicle_id] >= self.config["horizon"] and not self.is_multi_agent:
            done = True
            done_info[TerminationState.MAX_STEP] = True
            logging.info(msg("max step"))

        elif self.config["allowed_more_steps"] is not None and \
                self.episode_lengths[vehicle_id] >= self.engine.data_manager.current_scenario_length + self.config[
            "allowed_more_steps"] and not self.is_multi_agent:
            done = True
            done_info[TerminationState.MAX_STEP] = True
            logging.info(msg("more step than original episode"))

        # for compatibility
        # crash almost equals to crashing with vehicles
        done_info[TerminationState.CRASH] = (
            done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.CRASH_OBJECT]
            or done_info[TerminationState.CRASH_BUILDING]
        )

        # log data to curriculum manager
        self.engine.curriculum_manager.log_episode(done_info[TerminationState.SUCCESS], route_completion)

        return done, done_info

    def cost_function(self, vehicle_id: str):
        vehicle = self.vehicles[vehicle_id]
        step_info = dict(num_crash_object=0, num_crash_human=0, num_crash_vehicle=0, num_on_line=0)
        step_info["cost"] = 0
        if vehicle.on_yellow_continuous_line or vehicle.crash_sidewalk or vehicle.on_white_continuous_line:
            # step_info["cost"] += self.config["out_of_road_cost"]
            step_info["num_on_line"] = 1
        if self._is_out_of_road(vehicle):
            step_info["cost"] += self.config["out_of_road_cost"]
        if vehicle.crash_vehicle:
            step_info["cost"] += self.config["crash_vehicle_cost"]
            step_info["crash_vehicle_cost"] = self.config["crash_vehicle_cost"]
            step_info["num_crash_vehicle"] = 1
        if vehicle.crash_object:
            step_info["cost"] += self.config["crash_object_cost"]
            step_info["num_crash_object"] = 1
        if vehicle.crash_human:
            step_info["cost"] += self.config["crash_human_cost"]
            step_info["num_crash_human"] = 1
        return step_info['cost'], step_info

    def reward_function(self, vehicle_id: str):
        """
        Override this func to get a new reward function
        :param vehicle_id: id of BaseVehicle
        :return: reward
        """
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()

        # Reward for moving forward in current lane
        current_lane = vehicle.lane
        long_last = vehicle.navigation.last_longitude
        long_now = vehicle.navigation.current_longitude
        lateral_now = vehicle.navigation.current_lateral

        # dense driving reward
        reward = 0
        reward += self.config["driving_reward"] * (long_now - long_last)

        # reward for lane keeping, without it vehicle can learn to overtake but fail to keep in lane
        lateral_factor = abs(lateral_now) / self.config["max_lateral_dist"]
        lateral_penalty = -lateral_factor * self.config["lateral_penalty"]
        reward += lateral_penalty

        # heading diff
        ref_line_heading = vehicle.navigation.current_heading_theta_at_long
        heading_diff = wrap_to_pi(abs(vehicle.heading_theta - ref_line_heading)) / np.pi
        heading_penalty = -heading_diff * self.config["heading_penalty"]
        reward += heading_penalty

        # steering_range
        steering = abs(vehicle.current_action[0])
        allowed_steering = (1 / max(vehicle.speed, 1e-2))
        overflowed_steering = min((allowed_steering - steering), 0)
        steering_range_penalty = overflowed_steering * self.config["steering_range_penalty"]
        reward += steering_range_penalty

        if self.config["no_negative_reward"]:
            reward = max(reward, 0)

        # crash penalty
        if vehicle.crash_vehicle:
            reward = -self.config["crash_vehicle_penalty"]
        if vehicle.crash_object:
            reward = -self.config["crash_object_penalty"]
        if vehicle.crash_human:
            reward = -self.config["crash_human_penalty"]
        # lane line penalty
        if vehicle.on_yellow_continuous_line or vehicle.crash_sidewalk or vehicle.on_white_continuous_line:
            reward = -self.config["on_lane_line_penalty"]

        step_info["step_reward"] = reward

        # termination reward
        if self._is_arrive_destination(vehicle):
            reward = self.config["success_reward"]
        elif self._is_out_of_road(vehicle):
            reward = -self.config["out_of_road_penalty"]

        # TODO LQY: all a callback to process these keys
        step_info["track_length"] = vehicle.navigation.reference_trajectory.length
        step_info["carsize"] = [vehicle.WIDTH, vehicle.LENGTH]
        # add some new and informative keys
        step_info["route_completion"] = vehicle.navigation.route_completion
        step_info["curriculum_level"] = self.engine.current_level
        step_info["scenario_index"] = self.engine.current_seed
        step_info["num_stored_maps"] = self.engine.map_manager.num_stored_maps
        step_info["scenario_difficulty"] = self.engine.data_manager.current_scenario_difficulty
        step_info["data_coverage"] = self.engine.data_manager.data_coverage
        step_info["curriculum_success"] = self.engine.curriculum_manager.current_success_rate
        step_info["curriculum_route_completion"] = self.engine.curriculum_manager.current_route_completion
        step_info["lateral_dist"] = lateral_now

        step_info["step_reward_lateral"] = lateral_penalty
        step_info["step_reward_heading"] = heading_penalty
        step_info["step_reward_action_smooth"] = steering_range_penalty

        # Compute state difference metrics
        # TODO LQY: Shall we use state difference as reward?
        # data = self.engine.data_manager.current_scenario
        # agent_xy = vehicle.position
        # if vehicle_id == "sdc" or vehicle_id == "default_agent":
        #     native_vid = data[ScenarioDescription.METADATA][ScenarioDescription.SDC_ID]
        # else:
        #     native_vid = vehicle_id
        #
        # if native_vid in data["tracks"] and len(data["tracks"][native_vid]) > 0:
        #     expert_state_list = data["tracks"][native_vid]["state"]
        #
        #     mask = expert_state_list["valid"]
        #     largest_valid_index = np.max(np.where(mask == True)[0])
        #
        #     if self.episode_step > largest_valid_index:
        #         current_step = largest_valid_index
        #     else:
        #         current_step = self.episode_step
        #
        #     while mask[current_step] == 0.0:
        #         current_step -= 1
        #         if current_step == 0:
        #             break
        #
        #     expert_xy = expert_state_list["position"][current_step][:2]
        #     diff = agent_xy - expert_xy
        #     dist = norm(diff[0], diff[1])
        #     step_info["distance_error"] = dist
        #
        #     last_state = expert_state_list["position"][largest_valid_index]
        #     last_expert_xy = last_state[:2]
        #     diff = agent_xy - last_expert_xy
        #     last_dist = norm(diff[0], diff[1])
        #     step_info["distance_error_final"] = last_dist

        # reward = reward - self.config["distance_penalty"] * dist

        # if hasattr(vehicle, "_dynamics_mode"):
        #     step_info["dynamics_mode"] = vehicle._dynamics_mode

        return reward, step_info

    def _is_arrive_destination(self, vehicle):
        # Use RC as the only criterion to determine arrival in Scenario env.
        route_completion = vehicle.navigation.route_completion
        if route_completion > 0.95 or vehicle.navigation.reference_trajectory.length < 2:
            # Route Completion ~= 1.0 or vehicle is static!
            return True
        else:
            return False

    def _is_out_of_road(self, vehicle):
        # A specified function to determine whether this vehicle should be done.
        if self.config["relax_out_of_road_done"]:
            # We prefer using this out of road termination criterion.
            lat = abs(vehicle.navigation.current_lateral)
            done = lat > self.config["max_lateral_dist"]
            done = done
            return done

        done = vehicle.crash_sidewalk or vehicle.on_yellow_continuous_line or vehicle.on_white_continuous_line
        if self.config["out_of_route_done"]:
            done = done or abs(vehicle.navigation.current_lateral) > 10
        return done

    def _reset_global_seed(self, force_seed=None):
        if force_seed is not None:
            current_seed = force_seed
        elif self.config["sequential_seed"]:
            current_seed = self.engine.global_seed
            if current_seed is None:
                current_seed = int(self.config["start_scenario_index"]) + int(self.config["worker_index"])
            else:
                current_seed += int(self.config["num_workers"])
            if current_seed >= self.config["start_scenario_index"] + int(self.config["num_scenarios"]):
                current_seed = int(self.config["start_scenario_index"]) + int(self.config["worker_index"])
        else:
            current_seed = get_np_random(None).randint(
                self.config["start_scenario_index"],
                self.config["start_scenario_index"] + int(self.config["num_scenarios"])
            )

        assert self.config["start_scenario_index"] <= current_seed < self.config["start_scenario_index"] + \
               self.config["num_scenarios"], "Scenario Index (force seed) {} is out of range [{}, {}).".format(
            current_seed, self.config["start_scenario_index"],
            self.config["start_scenario_index"] + self.config["num_scenarios"])
        self.seed(current_seed)

    def stop(self):
        self.in_stop = not self.in_stop


if __name__ == "__main__":
    env = ScenarioEnv(
        {
            "use_render": True,
            "agent_policy": ReplayEgoCarPolicy,
            "manual_control": False,
            "show_interface": True,
            "show_logo": False,
            "show_fps": False,
            # "debug": True,
            # "debug_static_world": True,
            # "no_traffic": True,
            # "no_light": True,
            # "debug":True,
            # "no_traffic":True,
            # "start_scenario_index": 192,
            # "start_scenario_index": 1000,
            "num_scenarios": 30,
            # "force_reuse_object_name": True,
            # "data_directory": "/home/shady/Downloads/test_processed",
            "horizon": 1000,
            "no_static_vehicles": True,
            # "show_policy_mark": True,
            # "show_coordinates": True,
            "vehicle_config": dict(
                show_navi_mark=False,
                no_wheel_friction=True,
                lidar=dict(num_lasers=120, distance=50, num_others=4),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50)
            ),
            "data_directory": AssetLoader.file_path("nuplan", return_raw_style=False),
        }
    )
    success = []
    env.reset(seed=0)
    while True:
        env.reset(seed=env.current_seed + 1)
        for t in range(10000):
            o, r, tm, tc, info = env.step([0, 0])
            assert env.observation_space.contains(o)
            c_lane = env.vehicle.lane
            long, lat, = c_lane.local_coordinates(env.vehicle.position)
            # if env.config["use_render"]:
            env.render(
                text={
                    # "obs_shape": len(o),
                    "seed": env.engine.global_seed + env.config["start_scenario_index"],
                    # "reward": r,
                }
                # mode="topdown"
            )

            if (tm or tc) and info["arrive_dest"]:
                print("seed:{}, success".format(env.engine.global_random_seed))
                break



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

from panda3d.core import Camera, OrthographicLens, PTALVecBase3f

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.render_stage import RenderStage


class SkyAOCaptureStage(RenderStage):
    """ This stage captures the sky ao by rendering the scene from above """

    required_inputs = []
    required_pipes = []

    @property
    def produced_pipes(self):
        return {"SkyAOHeight": self.target_convert.color_tex}

    @property
    def produced_inputs(self):
        return {"SkyAOCapturePosition": self.pta_position}

    def __init__(self, pipeline):
        RenderStage.__init__(self, pipeline)
        self.pta_position = PTALVecBase3f.empty_array(1)
        self.resolution = 512
        self.capture_height = 100.0
        self.max_radius = 100.0

    def create(self):
        self.camera = Camera("SkyAOCaptureCam")
        self.cam_lens = OrthographicLens()
        self.cam_lens.set_film_size(self.max_radius, self.max_radius)
        self.cam_lens.set_near_far(0, self.capture_height)
        self.camera.set_lens(self.cam_lens)
        self.cam_node = Globals.base.render.attach_new_node(self.camera)
        self.cam_node.look_at(0, 0, -1)
        self.cam_node.set_r(0)

        self.target = self.create_target("SkyAOCapture")
        self.target.size = self.resolution
        self.target.add_depth_attachment(bits=16)
        self.target.prepare_render(self.cam_node)

        self.target_convert = self.create_target("ConvertDepth")
        self.target_convert.size = self.resolution
        self.target_convert.add_color_attachment(bits=(16, 0, 0, 0))
        self.target_convert.prepare_buffer()

        self.target_convert.set_shader_inputs(DepthTex=self.target.depth_tex, position=self.pta_position)

        # Register camera
        self._pipeline.tag_mgr.register_camera("shadow", self.camera)

    def update(self):
        snap_size = self.max_radius / self.resolution
        cam_pos = Globals.base.camera.get_pos(Globals.base.render)
        self.cam_node.set_pos(
            cam_pos.x - cam_pos.x % snap_size, cam_pos.y - cam_pos.y % snap_size, self.capture_height / 2.0
        )
        self.pta_position[0] = self.cam_node.get_pos()

    def reload_shaders(self):
        self.target_convert.shader = self.load_plugin_shader("convert_depth.frag.glsl")



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error


class PointerSlotStorage(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self, max_size):
        self._data = [None] * max_size
        self._max_index = 0
        self._num_entries = 0

    def get_max_index(self):
        return self._max_index

    def get_num_entries(self):
        return self._num_entries

    def find_slot(self):
        # Notice: returns None in case of no free slot, and the slot otherwise, this
        # is different to the C++ Module
        for i, value in enumerate(self._data):
            if not value:
                return i
        return -1

    def find_consecutive_slots(self, num_consecutive):
        if num_consecutive == 1:
            return self.find_slot()

        for i in range(len(self._data)):
            any_taken = False
            for k in range(num_consecutive):
                if self._data[i + k]:
                    any_taken = True
                    break
            if not any_taken:
                return i
        return -1

    def free_slot(self, slot):
        self._data[slot] = None
        self._num_entries -= 1
        if slot == self._max_index:
            while self._max_index >= 0 and not self._data[self._max_index]:
                self._max_index -= 1

    def free_consecutive_slots(self, slot, num_consecutive):
        for i in range(num_consecutive):
            self.free_slot(slot + i)

    def reserve_slot(self, slot, ptr):
        self._max_index = max(self._max_index, slot)
        self._data[slot] = ptr
        self._num_entries += 1

    def begin(self):
        for i in range(self._max_index + 1):
            if self._data[i]:
                yield self._data[i]

    def end(self):
        raise NotImplementedError("Use .begin() as iterator when using the python side!")



__all__ = ['CBaseLoader', 'CSafeLoader', 'CLoader', 'CBaseDumper', 'CSafeDumper', 'CDumper']

from yaml._yaml import CParser, CEmitter

from constructor import *

from serializer import *
from representer import *

from resolver import *


class CBaseLoader(CParser, BaseConstructor, BaseResolver):
    def __init__(self, stream):
        CParser.__init__(self, stream)
        BaseConstructor.__init__(self)
        BaseResolver.__init__(self)


class CSafeLoader(CParser, SafeConstructor, Resolver):
    def __init__(self, stream):
        CParser.__init__(self, stream)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)


class CLoader(CParser, Constructor, Resolver):
    def __init__(self, stream):
        CParser.__init__(self, stream)
        Constructor.__init__(self)
        Resolver.__init__(self)


class CBaseDumper(CEmitter, BaseRepresenter, BaseResolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=None,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None
    ):
        CEmitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            encoding=encoding,
            allow_unicode=allow_unicode,
            line_break=line_break,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags
        )
        Representer.__init__(self, default_style=default_style, default_flow_style=default_flow_style)
        Resolver.__init__(self)


class CSafeDumper(CEmitter, SafeRepresenter, Resolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=None,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None
    ):
        CEmitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            encoding=encoding,
            allow_unicode=allow_unicode,
            line_break=line_break,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags
        )
        SafeRepresenter.__init__(self, default_style=default_style, default_flow_style=default_flow_style)
        Resolver.__init__(self)


class CDumper(CEmitter, Serializer, Representer, Resolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=None,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None
    ):
        CEmitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            encoding=encoding,
            allow_unicode=allow_unicode,
            line_break=line_break,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags
        )
        Representer.__init__(self, default_style=default_style, default_flow_style=default_flow_style)
        Resolver.__init__(self)



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "start_seed": 4,
            "manual_control": True,
            "use_render": True,
            "image_observation": True,
            "rgb_clip": True,
            "vehicle_config": dict(depth_camera=(200, 88), image_source="depth_camera"),
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_NUM,
                BaseMap.GENERATE_CONFIG: 12,
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            }
        }
    )
    env.reset()
    env.engine.accept("m", env.vehicle.get_camera("depth_camera").save_image, extraArgs=[env.vehicle, "debug.jpg"])

    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        assert env.observation_space.contains(o)
        if env.config["use_render"]:
            # for i in range(ImageObservation.STACK_SIZE):
            #     ObservationType.show_gray_scale_array(o["image"][:, :, i])
            env.render(text={"can you see me": i})
        if tm or tc:
            env.vehicle.get_camera("rgb_camera").save_image(env.vehicle)
            # print("Reset")
            env.reset()
    env.close()



from panda3d.core import OmniBoundingVolume, Texture, Mat4, SamplerState, Vec3, PTALMatrix4f, PTAVecBase3f

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.water.water_manager import WaterManager


class ProjectedWater:
    def __init__(self, water_options):
        self.water_options = water_options
        self.water_level = 0.0
        self.model = Globals.base.loader.load_model("/$$rp/data/builtin_models/water/water_grid.bam")
        self.model.reparent_to(Globals.base.render)
        self.model.node().set_final(True)
        self.model.node().set_bounds(OmniBoundingVolume())
        self.model.set_two_sided(True)
        self.model.set_shader_input("waterHeight", self.water_level)
        self.model.set_mat(Mat4.identMat())
        self.model.clear_transform()

    def setup_water(self, pipeline, water_level):
        if (pipeline and self.model and isinstance(water_level, float) and self.water_options):

            self.pipeline = pipeline
            foam = Globals.base.loader.load_texture("/$$rp/data/builtin_models/water/water_foam.png")
            self.model.set_shader_input("waterFoam", foam)

            self.manager = WaterManager(self.water_options)
            self.manager.setup()
            self.manager.update()

            self.model.set_shader_input("waterHeightfield", self.manager.get_displacement_texture())
            self.model.set_shader_input("waterNormal", self.manager.get_normal_texture())

            # Set texture filter modes
            for tex in [foam, self.manager.get_displacement_texture(), self.manager.get_normal_texture()]:
                tex.set_wrap_u(SamplerState.WMRepeat)
                tex.set_wrap_u(SamplerState.WMRepeat)
                tex.set_minfilter(SamplerState.FTLinearMipmapLinear)
                tex.set_magfilter(SamplerState.FTLinearMipmapLinear)

            self.pipeline.set_effect(self.model, "/$$rp/effects/projected_water.yaml", {"render_shadow": False})

            self.water_level = water_level
            self.model.set_shader_input("waterHeight", self.water_level)

            Globals.base.add_task(self.update_task, "update_water")

    def clear_water(self):
        Globals.base.remove_task("update_water")
        if self.model:
            self.model.remove_node()

    def update_task(self, task):
        # Update water displacement
        self.manager.update()
        # Update the camera position and the current model view projection matrix
        self.model.set_shader_input("cameraPosition", Globals.base.camera.get_pos(render))
        self.model.set_shader_input("currentMVP", self.model.get_mat())
        return task.cont



from metadrive.envs.metadrive_env import MetaDriveEnv


def test_traffic_mode(render=False):
    try:
        for mode in ["hybrid", "trigger", "respawn"]:
            env = MetaDriveEnv(
                {
                    "num_scenarios": 1,
                    "traffic_density": 0.1,
                    "traffic_mode": mode,
                    "start_seed": 22,
                    "use_render": render,
                    "map": "X",
                }
            )

            o, _ = env.reset()
            env.vehicle.set_velocity([1, 0.1], 10)
            if mode == "respawn":
                assert len(env.engine.traffic_manager._traffic_vehicles) != 0
            elif mode == "hybrid" or mode == "trigger":
                assert len(env.engine.traffic_manager._traffic_vehicles) == 0
            for s in range(1, 300):
                o, r, tm, tc, info = env.step(env.action_space.sample())
            if mode == "hybrid" or mode == "respawn":
                assert len(env.engine.traffic_manager._traffic_vehicles) != 0
            elif mode == "trigger":
                assert len(env.engine.traffic_manager._traffic_vehicles) == 0
            env.close()
    finally:
        env.close()


if __name__ == "__main__":
    test_traffic_mode()



from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    config = {
        "num_scenarios": 10,
        "traffic_density": .0,
        # "use_render":True,
        "map": "SSSSS",
        # "manual_control":True,
        "controller": "joystick",
        "random_agent_model": False,
        "vehicle_config": {
            "vehicle_model": "default",
            # "vehicle_model":"s",
            # "vehicle_model":"m",
            # "vehicle_model":"l",
            # "vehicle_model":"xl",
        }
    }
    env = MetaDriveEnv(config)
    import time

    start = time.time()
    o, _ = env.reset()
    a = [.0, 1.]
    for s in range(1, 100000):
        o, r, tm, tc, info = env.step(a)
        if env.vehicle.speed_km_h > 100:
            a = [0, -1]
            # print("0-100 km/h acc use time:{}".format(s * 0.1))
            pre_pos = env.vehicle.position[0]
        if a == [0, -1] and env.vehicle.speed_km_h < 1:
            # print("0-100 brake use dist:{}".format(env.vehicle.position[0] - pre_pos))
            break
    env.close()



#!/usr/bin/env python3
# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse

from metadrive.third_party.kitsunetsuki.cardmaker import CardMaker

from panda3d.core import Filename, VirtualFileSystem, get_model_path


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--frames', type=str, help='comma separated frame groups')
    parser.add_argument('--fps', type=int, help='frames per second rate')
    parser.add_argument('--scale', type=float, help='object scale')
    parser.add_argument('--output', type=str, help='output file')
    parser.add_argument('--type', type=str, help='spritesheet file type')
    parser.add_argument('--input', type=str, help='input files', nargs='+')
    parser.add_argument('--empty', type=str, help='empty frame file')
    parser.add_argument('--prefix', type=str, help='texture path prefix')
    return parser.parse_args()


def main():
    args = parse_args()

    if args.frames:
        animations_frames = map(int, args.frames.split(','))
    else:
        animations_frames = [len(args.input)]

    kwargs = {}
    if args.fps:
        kwargs['fps'] = args.fps
    if args.scale:
        kwargs['scale'] = args.scale
    if args.type:
        kwargs['type'] = args.type
    if args.empty:
        kwargs['empty'] = args.empty
    if args.prefix:
        kwargs['prefix'] = args.prefix

    if args.prefix:
        vfs = VirtualFileSystem.get_global_ptr()
        vfs.mount(Filename.from_os_specific('.').get_fullpath(), args.prefix.rstrip('/'), 0)
        mp = get_model_path()
        mp.prepend_directory(args.prefix.rstrip('/'))

    cm = CardMaker(animations_frames, args.input, **kwargs)
    cm.make(args.output)


if __name__ == '__main__':
    main()



try:
    import inspect
    from typing import Any, Dict, Callable
    import gymnasium
    import gym
    import gym.spaces

    def gymnasiumToGym(space: gymnasium.spaces.Space) -> gym.spaces.Space:
        if isinstance(space, gymnasium.spaces.Box):
            return gym.spaces.Box(low=space.low, high=space.high, shape=space.shape)
        elif isinstance(space, gymnasium.spaces.Discrete):
            return gym.spaces.Discrete(n=int(space.n), start=int(space.start))
        elif isinstance(space, gymnasium.spaces.MultiDiscrete):
            return gym.spaces.MultiDiscrete(nvec=space.nvec)
        elif isinstance(space, gymnasium.spaces.Tuple):
            return gym.spaces.Tuple([gymnasiumToGym(subspace) for subspace in space.spaces])
        elif isinstance(space, gymnasium.spaces.Dict):
            return gym.spaces.Dict({key: gymnasiumToGym(subspace) for key, subspace in space.spaces.items()})
        else:
            raise ValueError("unsupported space")

    def gymToGymnasium(space: gym.spaces.Space) -> gymnasium.spaces.Space:
        if isinstance(space, gym.spaces.Box):
            return gymnasium.spaces.Box(low=space.low, high=space.high, shape=space.shape)
        elif isinstance(space, gym.spaces.Discrete):
            return gymnasium.spaces.Discrete(n=int(space.n), start=int(space.start))
        elif isinstance(space, gym.spaces.MultiDiscrete):
            return gymnasium.spaces.MultiDiscrete(nvec=space.nvec)
        elif isinstance(space, gym.spaces.Tuple):
            return gymnasium.spaces.Tuple([gymToGymnasium(subspace) for subspace in space.spaces])
        elif isinstance(space, gym.spaces.Dict):
            return gymnasium.spaces.Dict({key: gymToGymnasium(subspace) for key, subspace in space.spaces.items()})
        else:
            raise ValueError("unsupported space")

    def createGymWrapper(inner_class: type):
        """
        "inner_class": A gymnasium based Metadrive environment class
        """
        def was_overriden(a):
            """
            Returns if function `a` was not defined in this file
            """
            # we know that the function is overriden if the function was not defined in this file
            # because this is a dynamic class, equality checks will always return false
            # TODO: this is a hack, but i'm not sure how to make it more robust
            return inspect.getfile(a) != inspect.getfile(createGymWrapper)

        def createOverridenDefaultConfigWrapper(base: type, new_default_config: Callable) -> type:
            """
            Returns a class derived from the `base` class. It overrides the `default_config` classmethod, which is set to new_default_config
            """
            class OverridenDefaultConfigWrapper(base):
                @classmethod
                def default_config(cls):
                    return new_default_config()

            return OverridenDefaultConfigWrapper

        class GymEnvWrapper(gym.Env):
            @classmethod
            def default_config(cls):
                """
                This is the default, if you override it, then we will override it within the inner_class to maintain consistency
                """
                return inner_class.default_config()

            def __init__(self, config: Dict[str, Any]):
                # We can only tell if someone has overriden the default config method at init time.

                # if there was an override, we need to provide the overriden method to the inner class.
                if was_overriden(type(self).default_config):
                    # when inner_class's init is called, it now has access to the new default_config
                    actual_inner_class = createOverridenDefaultConfigWrapper(inner_class, type(self).default_config)
                else:
                    # otherwise, don't make a wrapper
                    actual_inner_class = inner_class

                # initialize
                super().__setattr__("_inner", actual_inner_class(config=config))

            def step(self, actions):
                o, r, tm, tc, i = self._inner.step(actions)
                if isinstance(tm, dict) and isinstance(tc, dict):
                    d = {j: (j in tm and tm[j]) or (j in tc and tc[j]) for j in set(list(tm.keys()) + list(tc.keys()))}
                else:
                    d = tm or tc
                return o, r, d, i

            def reset(self, *, seed=None, options=None):
                # pass non-none parameters to the reset (which may not support options or seed)
                params = {"seed": seed, "options": options}
                not_none_params = {k: v for k, v in params.items() if v is not None}
                obs, _ = self._inner.reset(**not_none_params)
                return obs

            def render(self, *args, **kwargs):
                # remove mode from kwargs
                kwargs.pop("mode", None)
                return self._inner.render(*args, **kwargs)

            def close(self):
                self._inner.close()

            def seed(self, seed=None):
                """
                We cannot seed a Gymnasium environment while running, so do nothing
                """
                pass

            @property
            def observation_space(self):
                return gymnasiumToGym(self._inner.observation_space)

            @property
            def action_space(self):
                return gymnasiumToGym(self._inner.action_space)

            def __getattr__(self, __name: str) -> Any:
                return getattr(self._inner, __name)

            def __setattr__(self, name, value):
                if hasattr(self._inner, name):
                    setattr(self._inner, name, value)
                else:
                    super().__setattr__(name, value)

        return GymEnvWrapper

    if __name__ == '__main__':
        from metadrive.envs.scenario_env import ScenarioEnv

        env = createGymWrapper(ScenarioEnv)(config={"manual_control": True})
        o, i = env.reset()
        assert isinstance(env.observation_space, gymnasium.spaces.Space)
        assert isinstance(env.action_space, gymnasium.spaces.Space)
        for s in range(600):
            o, r, d, i = env.step([0, -1])
            env.vehicle.set_velocity([0, 0])
            if d:
                assert s == env.config["horizon"] and i["max_step"] and d
                break
except:
    pass



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def uv_equals(uv1, uv2):
    u1, v1 = uv1
    u2, v2 = uv2

    if abs(u1 - u2) > 0.001:
        return False

    if abs(v1 - v2) > 0.001:
        return False

    return True


def normal_equals(n1, n2):
    x1, y1, z1 = n1
    x2, y2, z2 = n2

    if abs(x1 - x2) > 0.001:
        return False

    if abs(y1 - y2) > 0.001:
        return False

    if abs(z1 - z2) > 0.001:
        return False

    return True



from typing import Set

import math
import numpy as np
from panda3d.bullet import BulletGhostNode, BulletCylinderShape
from panda3d.core import NodePath

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.sensors.distance_detector import DistanceDetector
from metadrive.constants import CamMask, CollisionGroup
from metadrive.engine.engine_utils import get_engine
from metadrive.utils.coordinates_shift import panda_vector
from metadrive.utils.math import norm, clip
from metadrive.utils.utils import get_object_from_node


class Lidar(DistanceDetector):
    ANGLE_FACTOR = True
    Lidar_point_cloud_obs_dim = 240
    DEFAULT_HEIGHT = 1.2

    BROAD_PHASE_EXTRA_DIST = 0

    def __init__(self, num_lasers: int = 240, distance: float = 50, enable_show=False):
        super(Lidar, self).__init__(num_lasers, distance, enable_show)
        self.origin.hide(CamMask.RgbCam | CamMask.Shadow | CamMask.Shadow | CamMask.DepthCam)
        self.mask = CollisionGroup.can_be_lidar_detected()

        # lidar can calculate the detector mask by itself
        self.angle_delta = 360 / num_lasers if num_lasers > 0 else None
        self.broad_detector = NodePath(BulletGhostNode("detector_mask"))
        self.broad_detector.node().addShape(BulletCylinderShape(self.BROAD_PHASE_EXTRA_DIST + distance, 5))
        self.broad_detector.node().setIntoCollideMask(CollisionGroup.LidarBroadDetector)
        self.broad_detector.node().setStatic(True)
        engine = get_engine()
        engine.physics_world.static_world.attach(self.broad_detector.node())
        self.enable_mask = True if not engine.global_config["_disable_detector_mask"] else False

        self._node_path_list.append(self.broad_detector)

    def perceive(self, base_vehicle, detector_mask=True):
        res = self._get_lidar_mask(base_vehicle)
        lidar_mask = res[0] if detector_mask and self.enable_mask else None
        detected_objects = res[1]
        return super(Lidar, self).perceive(base_vehicle, base_vehicle.engine.physics_world.dynamic_world,
                                           lidar_mask)[0], detected_objects

    @staticmethod
    def get_surrounding_vehicles(detected_objects) -> Set:
        from metadrive.component.vehicle.base_vehicle import BaseVehicle
        vehicles = set()
        objs = detected_objects
        for ret in objs:
            if isinstance(ret, BaseVehicle):
                vehicles.add(ret)
        return vehicles

    def _project_to_vehicle_system(self, target, vehicle):
        diff = target - vehicle.position
        norm_distance = norm(diff[0], diff[1])
        if norm_distance > self.perceive_distance:
            diff = diff / norm_distance * self.perceive_distance
        relative = vehicle.convert_to_local_coordinates(diff, 0.0)
        return relative

    def get_surrounding_vehicles_info(self, ego_vehicle, detected_objects, num_others: int = 4, add_others_navi=False):
        surrounding_vehicles = list(self.get_surrounding_vehicles(detected_objects))
        surrounding_vehicles.sort(
            key=lambda v: norm(ego_vehicle.position[0] - v.position[0], ego_vehicle.position[1] - v.position[1])
        )
        surrounding_vehicles += [None] * num_others
        res = []

        for vehicle in surrounding_vehicles[:num_others]:
            if vehicle is not None:
                ego_position = ego_vehicle.position

                # assert isinstance(vehicle, IDMVehicle or Base), "Now MetaDrive Doesn't support other vehicle type"
                relative_position = ego_vehicle.convert_to_local_coordinates(vehicle.position, ego_position)
                # It is possible that the centroid of other vehicle is too far away from ego but lidar shed on it.
                # So the distance may greater than perceive distance.
                res.append(clip((relative_position[0] / self.perceive_distance + 1) / 2, 0.0, 1.0))
                res.append(clip((relative_position[1] / self.perceive_distance + 1) / 2, 0.0, 1.0))

                relative_velocity = ego_vehicle.convert_to_local_coordinates(
                    vehicle.velocity_km_h, ego_vehicle.velocity_km_h
                )
                res.append(clip((relative_velocity[0] / ego_vehicle.max_speed_km_h + 1) / 2, 0.0, 1.0))
                res.append(clip((relative_velocity[1] / ego_vehicle.max_speed_km_h + 1) / 2, 0.0, 1.0))

                if add_others_navi:
                    ckpt1, ckpt2 = vehicle.navigation.get_checkpoints()

                    relative_ckpt1 = self._project_to_vehicle_system(ckpt1, ego_vehicle)
                    res.append(clip((relative_ckpt1[0] / self.perceive_distance + 1) / 2, 0.0, 1.0))
                    res.append(clip((relative_ckpt1[1] / self.perceive_distance + 1) / 2, 0.0, 1.0))

                    relative_ckpt2 = self._project_to_vehicle_system(ckpt2, ego_vehicle)
                    res.append(clip((relative_ckpt2[0] / self.perceive_distance + 1) / 2, 0.0, 1.0))
                    res.append(clip((relative_ckpt2[1] / self.perceive_distance + 1) / 2, 0.0, 1.0))

            else:

                if add_others_navi:
                    res += [0.0] * 8
                else:
                    res += [0.0] * 4

        return res

    def _get_lidar_mask(self, vehicle):
        pos1 = vehicle.position
        head1 = vehicle.heading_theta

        mask = np.zeros((self.num_lasers, ), dtype=bool)
        mask.fill(False)
        objs = self.get_surrounding_objects(vehicle)
        for obj in objs:
            pos2 = obj.position
            length = obj.LENGTH if hasattr(obj, "LENGTH") else vehicle.LENGTH
            width = obj.WIDTH if hasattr(obj, "WIDTH") else vehicle.WIDTH
            half_max_span_square = ((length + width) / 2)**2
            diff = (pos2[0] - pos1[0], pos2[1] - pos1[1])
            dist_square = diff[0]**2 + diff[1]**2
            if dist_square < half_max_span_square:
                mask.fill(True)
                continue

            span = math.asin(math.sqrt(half_max_span_square / dist_square))
            # relative heading of v2's center when compared to v1's center
            relative_head = math.atan2(diff[1], diff[0])
            head_in_1 = relative_head - head1
            head_in_1_max = head_in_1 + span
            head_in_1_min = head_in_1 - span
            head_1_max = np.rad2deg(head_in_1_max)
            head_1_min = np.rad2deg(head_in_1_min)
            mask = self._mark_this_range(head_1_min, head_1_max, mask)

        return mask, objs

    def get_surrounding_objects(self, vehicle):
        self.broad_detector.setPos(panda_vector(vehicle.position))
        physics_world = vehicle.engine.physics_world.dynamic_world
        contact_results = physics_world.contactTest(self.broad_detector.node(), True).getContacts()
        objs = set()
        for contact in contact_results:
            node0 = contact.getNode0()
            node1 = contact.getNode1()
            nodes = [node0, node1]
            nodes.remove(self.broad_detector.node())
            obj = get_object_from_node(nodes[0])
            if not isinstance(obj, AbstractLane) and obj is not None:
                objs.add(obj)
        if vehicle in objs:
            objs.remove(vehicle)
        return objs

    def _mark_this_range(self, small_angle, large_angle, mask):
        # We use clockwise to determine small and large angle.
        # For example, if you wish to fill 355 deg to 5 deg, then small_angle is 355, large_angle is 5.
        small_angle = small_angle % 360
        large_angle = large_angle % 360

        assert 0 <= small_angle <= 360
        assert 0 <= large_angle <= 360

        small_index = math.floor(small_angle / self.angle_delta)
        large_index = math.ceil(large_angle / self.angle_delta)
        if large_angle < small_angle:  # We are in the case like small=355, large=5
            mask[small_index:] = True
            mask[:large_index + 1] = True
        else:
            mask[small_index:large_index + 1] = True
        return mask

    def destroy(self):
        get_engine().physics_world.static_world.remove(self.broad_detector.node())
        self.broad_detector.removeNode()
        super(Lidar, self).destroy()



"""
This script aims to convert nuscenes scenarios to ScenarioDescription, so that we can load any nuscenes scenarios into
MetaDrive.
"""
import copy
import os
import pickle

import tqdm

from metadrive.engine.asset_loader import AssetLoader
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.utils.nuscenes.utils import convert_one_scenario
from metadrive.utils.utils import dict_recursive_remove_array
import shutil
import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)

try:
    from nuscenes import NuScenes
except ImportError:
    print("Can not find nuscenes-devkit")


def convert_scenarios(version, dataroot, output_path, worker_index=None, verbose=True, force_overwrite=False):
    save_path = copy.deepcopy(output_path)
    output_path = output_path + "_tmp"
    # meta recorder and data summary
    if os.path.exists(output_path):
        shutil.rmtree(output_path)
    os.makedirs(output_path, exist_ok=False)

    # make real save dir
    delay_remove = None
    if os.path.exists(save_path):
        if force_overwrite:
            delay_remove = save_path
        else:
            raise ValueError("Directory already exists! Abort")

    metadata_recorder = {}
    total_scenarios = 0
    desc = ""
    summary_file = ScenarioDescription.DATASET.SUMMARY_FILE
    if worker_index is not None:
        desc += "Worker {} ".format(worker_index)
        summary_file = "dataset_summary_worker{}.pkl".format(worker_index)

    # Init.
    nusc = NuScenes(version=version, verbose=verbose, dataroot=dataroot)
    scenes = nusc.scene
    for scene in tqdm.tqdm(scenes):
        sd_scene = convert_one_scenario(scene["token"], nusc)
        sd_scene = sd_scene.to_dict()
        ScenarioDescription.sanity_check(sd_scene, check_self_type=True)
        export_file_name = ScenarioDescription.get_export_file_name("nuscenes", version, scene["token"])
        p = os.path.join(output_path, export_file_name)
        with open(p, "wb") as f:
            pickle.dump(sd_scene, f)
        metadata_recorder[export_file_name] = copy.deepcopy(sd_scene[ScenarioDescription.METADATA])
    # rename and save
    if delay_remove is not None:
        shutil.rmtree(delay_remove)
    os.rename(output_path, save_path)
    summary_file = os.path.join(save_path, summary_file)
    with open(summary_file, "wb") as file:
        pickle.dump(dict_recursive_remove_array(metadata_recorder), file)
    print("Summary is saved at: {}".format(summary_file))
    assert delay_remove == save_path


if __name__ == "__main__":
    output_path = AssetLoader.file_path("nuscenes", return_raw_style=False)
    version = 'v1.0-mini'
    verbose = True
    dataroot = '/home/shady/data/nuscenes'
    worker_index = None
    force_overwrite = True
    convert_scenarios(version, dataroot, output_path, force_overwrite=force_overwrite)



from metadrive import ActionRepeat
from metadrive.constants import TerminationState


def _test_action_repeat(config):
    env = ActionRepeat(config)
    try:
        env.reset()
        for i in range(100):
            _, _, d, info = env.step(env.action_space.sample())

            for k in ["simulation_time", "real_return", "action_repeat", "primitive_steps_count",
                      TerminationState.MAX_STEP, "render", "trajectory"]:
                assert k in info

            assert len(info["trajectory"]) == info["action_repeat"]
            for t in info["trajectory"]:
                for k in ["reward_function", "discounted_reward", "obs", "action", "count"]:
                    assert k in t

            if d:
                env.reset()
    finally:
        env.close()


def test_action_repeat_env():
    _test_action_repeat(dict(max_action_repeat=5))
    _test_action_repeat(dict(max_action_repeat=1))
    _test_action_repeat(dict(fixed_action_repeat=5))
    _test_action_repeat(dict(fixed_action_repeat=1))


if __name__ == '__main__':
    test_action_repeat_env()
    # pytest.main(["-sv", "_test_action_repeat_env.py"])



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import TransparencyAttrib, SamplerState
from direct.gui.DirectCheckBox import DirectCheckBox
import direct.gui.DirectGuiGlobals as DGG

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader


class Checkbox(RPObject):
    """ This is a wrapper around DirectCheckBox, providing a simpler interface
    and better visuals """
    def __init__(
        self,
        parent=None,
        x=0,
        y=0,
        callback=None,
        extra_args=None,
        radio=False,
        expand_width=100,
        checked=False,
        enabled=True
    ):
        RPObject.__init__(self)

        prefix = "checkbox" if not radio else "radiobox"

        if enabled:
            checked_img = RPLoader.load_texture("/$$rp/data/gui/" + prefix + "_checked.png")
            unchecked_img = RPLoader.load_texture("/$$rp/data/gui/" + prefix + "_default.png")
        else:
            checked_img = RPLoader.load_texture("/$$rp/data/gui/" + prefix + "_disabled.png")
            unchecked_img = checked_img

        # Set near filter, otherwise textures look like crap
        for tex in [checked_img, unchecked_img]:
            tex.set_minfilter(SamplerState.FT_linear)
            tex.set_magfilter(SamplerState.FT_linear)
            tex.set_wrap_u(SamplerState.WM_clamp)
            tex.set_wrap_v(SamplerState.WM_clamp)
            tex.set_anisotropic_degree(0)

        self._node = DirectCheckBox(
            parent=parent,
            pos=(x + 11, 1, -y - 8),
            scale=(10 / 2.0, 1, 10 / 2.0),
            checkedImage=checked_img,
            uncheckedImage=unchecked_img,
            image=unchecked_img,
            extraArgs=extra_args,
            state=DGG.NORMAL,
            relief=DGG.FLAT,
            command=self._update_status
        )

        self._node["frameColor"] = (0, 0, 0, 0)
        self._node["frameSize"] = (-2.6, 2 + expand_width / 7.5, -2.35, 2.5)
        self._node.set_transparency(TransparencyAttrib.M_alpha)

        self._callback = callback
        self._extra_args = extra_args
        self._collection = None

        if checked:
            self.set_checked(True, False)

    @property
    def collection(self):
        """ Returns a handle to the assigned checkbox collection, or None
        if no collection was assigned """
        return self._collection

    @collection.setter
    def collection(self, coll):
        """ Internal method to add a checkbox to a checkbox collection, this
        is used for radio-buttons """
        self._collection = coll

    @property
    def checked(self):
        """ Returns whether the node is currently checked """
        return self._node["isChecked"]

    @property
    def node(self):
        """ Returns a handle to the internally used node """
        return self._node

    def _update_status(self, status):
        """ Internal method when another checkbox in the same radio group
        changed it's value """

        if not status and self._collection:
            self._node.commandFunc(None)
            return

        if self._collection:
            if status:
                self._collection.on_checkbox_changed(self)
                # A radio box can't be unchecked
                # self._node["state"] = DGG.DISABLED

        if self._callback is not None:
            self._callback(*([status] + self._extra_args))

    def set_checked(self, val, do_callback=True):
        """ Internal method to check/uncheck the checkbox """
        self._node["isChecked"] = val

        if val:
            self._node["image"] = self._node["checkedImage"]
        else:
            self._node["image"] = self._node["uncheckedImage"]

        if do_callback and self._callback is not None:
            self._callback(*([val] + self._extra_args))



"""
This script demonstrates how to use the environment where traffic and road map are loaded from argoverse dataset.
"""
import random

from metadrive.component.vehicle_navigation_module.trajectory_navigation import WaymoTrajectoryNavigation
from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.replay_policy import WaymoReplayEgoCarPolicy


class DemoWaymoEnv(WaymoEnv):
    def reset(self, seed=None):
        if self.engine is not None and seed is None:
            seeds = [i for i in range(self.config["num_scenarios"])]
            seeds.remove(self.current_seed)
            seed = random.choice(seeds)
        return super(DemoWaymoEnv, self).reset(seed=seed)


if __name__ == "__main__":
    asset_path = AssetLoader.asset_path
    env = DemoWaymoEnv(
        {
            "manual_control": False,
            "agent_policy": WaymoReplayEgoCarPolicy,
            "use_render": True,
            "data_directory": AssetLoader.file_path(asset_path, "waymo", return_raw_style=False),
            "num_scenarios": 3,
            "start_scenario_index": 0,
            "crash_vehicle_done": False,
            "crash_vehicle_penalty": 0,
            "vehicle_config": {
                "navigation_module": WaymoTrajectoryNavigation,
                "show_side_detector": True,
                "show_lane_line_detector": True,
            }
        }
    )
    o, _ = env.reset(seed=0)

    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([1.0, 0.])
        # print(env.vehicle.height)
        env.render(text={"seed": env.current_seed, "reward": r})
        if tm or tc:
            # print(info["arrive_dest"])
            env.reset()
    env.close()



from collections import deque

from metadrive.engine.engine_utils import get_engine
import numpy as np


def _clear_if_necessary(obj, depth=0):
    if depth > 5:
        obj = None
        del obj
        return

    if isinstance(obj, dict):
        keys = set(obj.keys())
        for k in keys:
            if k in obj:
                _clear_if_necessary(obj[k], depth + 1)
                obj[k] = None
        obj.clear()
    elif isinstance(obj, str):
        del obj
    elif isinstance(obj, list):
        obj.clear()
        del obj
    elif isinstance(obj, np.ndarray):
        del obj
    else:
        del obj


class DataBuffer:
    """
    This class builds a cache for data (mainly for Waymo data).
    You can store map / traffic tracks in it.
    The maximum size of the buffer is determined by store_data_buffer_size
    """
    def __init__(self, store_data_buffer_size=None):

        self.store_data_buffer = {}

        if store_data_buffer_size is None:
            store_data_buffer_size = 1

        self.store_data_indices = deque(maxlen=store_data_buffer_size)
        self.store_data_buffer_size = store_data_buffer_size

    def clear_if_necessary(self):
        while len(self.store_data_buffer) >= self.store_data_buffer_size:

            tmp_index = self.store_data_indices.popleft()

            if not isinstance(self.store_data_buffer[tmp_index], dict):
                self.store_data_indices.appendleft(tmp_index)
                return

            tmp_obj = self.store_data_buffer.pop(tmp_index)
            get_engine().clear_object_if_possible(tmp_obj, force_destroy=True)
            if hasattr(tmp_obj, "destroy"):
                print("Destroy object: ", type(tmp_obj))
                tmp_obj.destroy()

            _clear_if_necessary(tmp_obj)

    def __getitem__(self, item):
        assert item in self.store_data_buffer
        return self.store_data_buffer[item]

    def __contains__(self, key):
        return key in self.store_data_buffer

    def __setitem__(self, key, value):
        # This should be checked in map manager instead of here
        # map_num = get_engine().global_config["num_scenarios"]
        # start = get_engine().global_config["start_scenario_index"]
        #
        # assert start <= key < start + map_num, (start, key, start + map_num)

        self.clear_if_necessary()

        self.store_data_buffer[key] = value
        self.store_data_indices.append(key)

        # assert len(self.store_data_indices) == len(self.store_data_buffer)
        # assert len(self.store_data_indices) <= self.store_data_buffer_size



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import math
from panda3d.core import Vec3

from metadrive.render_pipeline.rpcore.pynative.shadow_source import ShadowSource
from metadrive.render_pipeline.rpcore.pynative.rp_light import RPLight


class RPSpotLight(RPLight):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self):
        RPLight.__init__(self, RPLight.LT_spot_light)
        self._radius = 10.0
        self._fov = 45.0
        self._direction = Vec3(0, 0, -1)

    def write_to_command(self, cmd):
        RPLight.write_to_command(self, cmd)
        cmd.push_float(self._radius)
        cmd.push_float(math.cos(self._fov / 360.0 * math.pi))
        cmd.push_vec3(self._direction)

    def set_radius(self, radius):
        self._radius = radius
        self.set_needs_update(True)
        self.invalidate_shadows()

    def get_radius(self):
        return self._radius

    radius = property(get_radius, set_radius)

    def set_fov(self, fov):
        self._fov = fov
        self.set_needs_update(True)
        self.invalidate_shadows()

    def get_fov(self):
        return self._fov

    fov = property(get_fov, set_fov)

    def set_direction(self, *args):
        self._direction = Vec3(*args)
        self._direction.normalize()
        self.set_needs_update(True)
        self.invalidate_shadows()

    def get_direction(self):
        return self._direction

    direction = property(get_direction, set_direction)

    def look_at(self, *args):
        vec = Vec3(*args) - self._position
        self.set_direction(vec)

    def init_shadow_sources(self):
        self._shadow_sources.append(ShadowSource())

    def update_shadow_sources(self):
        self._shadow_sources[0].set_resolution(self.get_shadow_map_resolution())
        self._shadow_sources[0].set_perspective_lens(
            self._fov, self._near_plane, self._radius, self._position, self._direction
        )



from __future__ import division, print_function

from panda3d.core import PNMImage, Vec3


def approx(roughness):
    return 1.0 - 0.5 * roughness


for i in range(11):
    r = i / 10.0
    fname = "batch_compare/Gold-R" + str(r) + ".png"

    img = PNMImage(fname)
    # color = (img.get_xel(img.get_x_size() // 2 + 3, img.get_y_size() // 2))
    color = (img.get_xel(256, 47))

    color.x = pow(color.x, 2.2)
    color.y = pow(color.y, 2.2)
    color.z = pow(color.z, 2.2)

    # print(color)

    basecolor = Vec3(1, 0.867136, 0.358654)

    ref_r, apprx_r = color.x, approx(r) * (0.5 + 0.5 * basecolor.x)
    ref_g, apprx_g = color.y, approx(r) * (0.5 + 0.5 * basecolor.y)
    ref_b, apprx_b = color.z, approx(r) * (0.5 + 0.5 * basecolor.z)
    print("Roughness:", r, ", color = ", ref_r, ref_g, ref_b, "vs", apprx_r, apprx_g, apprx_b)
    l = abs(ref_r - apprx_r) + abs(ref_g - apprx_g) + abs(ref_b - apprx_b)
    print("  --> ", l)
    # print("Roughness:", r, ", color = ", ref, "vs", apprx, " =\t", abs(ref - apprx) * 100.0)



from panda3d.core import LVector4, NodePath, DirectionalLight, AmbientLight

from metadrive.base_class.base_object import BaseObject
from metadrive.constants import CamMask


class Light(BaseObject):
    """
    It is dynamic element since it will follow the camera
    """
    def __init__(self, config):
        super(Light, self).__init__(random_seed=0)
        self.global_light = config["global_light"]
        self.direction_np = NodePath(DirectionalLight("direction light"))
        # self.light.node().setScene(self.render)

        self._node_path_list.append(self.direction_np)

        # Too large will cause the graphics card out of memory.
        if self.global_light:
            self.direction_np.node().setShadowCaster(True, 8192, 8192)
            # self.direction_np.setPos(0, 0, 50)
            # self.direction_np.lookAt(100, -30, 0)
        else:
            self.direction_np.node().setShadowCaster(True, 128, 128)

        # self.direction_np.node().showFrustum()
        # self.light.node().getLens().setNearFar(10, 100)

        self.direction_np.node().setColor(LVector4(1, 1, 1, 1))
        self.direction_np.node().setCameraMask(CamMask.Shadow)

        dlens = self.direction_np.node().getLens()
        if self.global_light:
            dlens.setFilmSize(64, 64)
        else:
            dlens.setFilmSize(16, 16)
        # dlens.setFocalLength(1)
        # dlens.setNear(3)

        self.direction_np.node().setColorTemperature(6800)
        self.direction_np.reparentTo(self.origin)

        self.ambient_np = NodePath(AmbientLight("Ambient"))
        self.ambient_np.node().setColor(LVector4(0.18, 0.2, 0.2, 1))
        # self.ambient_np.node().setColor(LVector4(0.8, 0.8, 0.8, 1))
        self.ambient_np.reparentTo(self.origin)

        self._node_path_list.append(self.ambient_np)

    def step(self, pos):
        # if not self.global_light:
        self.direction_np.setPos(pos[0] - 100, pos[1] + 100, 120)
        self.direction_np.lookAt(pos[0], pos[1], 0)

    def reset(self, random_seed=None, *args, **kwargs):
        if self.direction_np is not None and not self.global_light:
            self.direction_np.setHpr(-90, -120, 0)
            self.direction_np.setY(20)
            self.direction_np.setX(0)
            self.direction_np.setZ(50)



"""
Script from https://github.com/metadriverse/metadrive/issues/290
Need to install zmq and lz4
"""
import zmq
import lz4.frame as lz4
import argparse
import random
from metadrive import MetaDriveEnv
from metadrive.constants import HELP_MESSAGE
import threading
import numpy as np
import cv2

W, H = 1920, 1200


def main_thread():
    config = dict(
        # controller="joystick",
        use_render=True,
        image_observation=True,
        manual_control=True,  # set false for external subscriber control
        traffic_density=0.0,
        num_scenarios=100,
        random_agent_model=True,
        random_lane_width=True,
        random_lane_num=True,
        vehicle_config=dict(image_source="rgb_camera", rgb_camera=(W, H), stack_size=1),
        map=4,  # seven block
        start_seed=random.randint(0, 1000),
        window_size=(300, 200)
    )
    env = MetaDriveEnv(config)
    try:
        o, _ = env.reset()
        # print(HELP_MESSAGE)
        env.vehicle.expert_takeover = False
        context = zmq.Context()
        socket = context.socket(zmq.PUSH)
        socket.bind("tcp://127.0.0.1:5555")
        socket.setsockopt(zmq.SNDHWM, 1)
        socket.setsockopt(zmq.RCVHWM, 1)
        vehicle_socket = context.socket(zmq.PUSH)
        vehicle_socket.bind("tcp://127.0.0.1:5556")  # vehicle state
        vehicle_socket.setsockopt(zmq.SNDHWM, 1)
        vehicle_socket.setsockopt(zmq.RCVHWM, 1)

        assert isinstance(o, dict)
        throttle_brake_op = 2.0
        steer_op = 0.0
        while True:
            # sm.update(0) # cereal submaster
            # throttle_brake_op = sm['carControl'].actuators.accel
            # steer_op = sm['carControl'].actuators.steeringAngleDeg
            o, rewards, dones, step_infos = env.step([steer_op, throttle_brake_op])
            try:
                img = o['image']  # (W,H,3,stack_size)
                compressed_image = lz4.compress(img)
                socket.send(compressed_image, zmq.NOBLOCK)
            except zmq.error.Again:
                # print("Dropped frame")
                pass
            try:
                vehicle_socket.send_pyobj(step_infos, zmq.NOBLOCK)
            except zmq.error.Again:
                # print("Dropped vehicle state")
                pass
    except Exception as e:
        raise e
    finally:
        env.close()
        socket.close()
        context.term()


if __name__ == "__main__":
    main_thread()



from typing import Tuple, Sequence

from panda3d.core import LVector3, NodePath

from metadrive.base_class.base_object import BaseObject
from metadrive.constants import CollisionGroup

LaneIndex = Tuple[str, str, int]


class BaseTrafficParticipant(BaseObject):
    TYPE_NAME = None
    COLLISION_MASK = CollisionGroup.TrafficParticipants
    HEIGHT = None

    def __init__(self, position: Sequence[float], heading_theta: float = 0., random_seed=None, name=None):
        super(BaseTrafficParticipant, self).__init__(random_seed=random_seed, name=name)
        self.set_position(position, self.HEIGHT / 2 if hasattr(self, "HEIGHT") else 0)
        self.set_heading_theta(heading_theta)
        assert self.MASS is not None, "No mass for {}".format(self.class_name)
        assert self.TYPE_NAME is not None, "No name for {}".format(self.class_name)
        assert self.COLLISION_MASK is not None, "No collision group for {}".format(self.class_name)

    def reset(self, position: Sequence[float], heading_theta: float = 0., random_seed=None, name=None, *args, **kwargs):
        self.seed(random_seed)
        self.rename(name)
        self.set_position(position, self.HEIGHT / 2 if hasattr(self, "HEIGHT") else 0)
        self.set_heading_theta(heading_theta)
        assert self.MASS is not None, "No mass for {}".format(self.class_name)
        assert self.TYPE_NAME is not None, "No name for {}".format(self.class_name)
        assert self.COLLISION_MASK is not None, "No collision group for {}".format(self.class_name)

    @property
    def top_down_width(self):
        return self.WIDTH

    @property
    def top_down_length(self):
        return self.LENGTH

    # def set_roll(self, roll):
    #     self.origin.setP(roll)
    #
    # def set_pitch(self, pitch):
    #     self.origin.setR(pitch)
    #
    # @property
    # def roll(self):
    #     return np.deg2rad(self.origin.getP())
    #
    # @property
    # def pitch(self):
    #     return np.deg2rad(self.origin.getR())

    def add_body(self, physics_body):
        super(BaseTrafficParticipant, self).add_body(physics_body)
        self._body.set_friction(0.)
        self._body.set_anisotropic_friction(LVector3(0., 0., 0.))

    def standup(self):
        self.set_pitch(0)
        self.set_roll(0)

    def set_position(self, position, height=None):
        super(BaseTrafficParticipant, self).set_position(position, height)
        self.standup()

    def show_coordinates(self):
        if not self.need_show_coordinates:
            return
        if self.coordinates_debug_np is not None:
            self.coordinates_debug_np.reparentTo(self.origin)
        height = self.HEIGHT
        self.coordinates_debug_np = NodePath("debug coordinate")
        x = self.engine.draw_line_3d([0, 0, height], [1, 0, height], [1, 0, 0, 1], 1)
        y = self.engine.draw_line_3d([0, 0, height], [0, 0.5, height], [1, 0, 0, 1], 1)
        z = self.engine.draw_line_3d([0, 0, height], [0, 0, height + 0.5], [0, 0, 1, 1], 1)
        x.reparentTo(self.coordinates_debug_np)
        y.reparentTo(self.coordinates_debug_np)
        z.reparentTo(self.coordinates_debug_np)
        self.coordinates_debug_np.reparentTo(self.origin)



import logging
import sys

import numpy as np

from metadrive.component.sensors.vehicle_panel import VehiclePanel
from metadrive.component.vehicle_navigation_module.trajectory_navigation import NuPlanTrajectoryNavigation
from metadrive.constants import TerminationState
from metadrive.envs.base_env import BaseEnv
from metadrive.manager.nuplan_data_manager import NuPlanDataManager
from metadrive.manager.nuplan_light_manager import NuPlanLightManager
from metadrive.manager.nuplan_map_manager import NuPlanMapManager
from metadrive.manager.nuplan_traffic_manager import NuPlanTrafficManager
from metadrive.obs.real_env_observation import NuPlanObservation
from metadrive.obs.state_obs import LidarStateObservation
from metadrive.policy.replay_policy import NuPlanReplayEgoCarPolicy
from metadrive.utils import clip
from metadrive.utils import get_np_random
from metadrive.utils.coordinates_shift import nuplan_to_metadrive_vector

NUPLAN_ENV_CONFIG = dict(
    # ===== Dataset Config =====
    # These parameters remain the same as the meaning of those in nuplan-devkit
    DATASET_PARAMS=[
        'scenario_builder=nuplan_mini',  # use nuplan mini database (2.5h of 8 autolabeled logs in Las Vegas)
        'scenario_filter=one_continuous_log',  # simulate only one log
        "scenario_filter.log_names=['2021.07.16.20.45.29_veh-35_01095_01486']",
        'scenario_filter.limit_total_scenarios=2800',  # use 2 total scenarios
    ],
    start_scenario_index=0,
    num_scenarios=100,
    store_map=True,
    sequential_seed=False,

    # ===== Map Config =====
    city_map_radius=20000,  # load the whole map, setting as large as possible
    scenario_radius=250,  # radius for per scenario
    load_city_map=False,
    map_centers={
        'us-nv-las-vegas-strip': nuplan_to_metadrive_vector([664396, 3997613]),
        'sg-one-north': nuplan_to_metadrive_vector([365427, 143908]),
        'us-pa-pittsburgh-hazelwood': nuplan_to_metadrive_vector([587631, 4475539])
    },

    # ===== Traffic =====
    no_pedestrian=True,
    no_traffic=False,
    no_light=False,
    reactive_traffic=False,

    # ===== Agent config =====
    vehicle_config=dict(
        lidar=dict(num_lasers=120, distance=50),
        lane_line_detector=dict(num_lasers=12, distance=50),
        side_detector=dict(num_lasers=120, distance=50),
        show_dest_mark=True,
        navigation_module=NuPlanTrajectoryNavigation,
    ),
    use_nuplan_observation=True,

    # ===== Reward Scheme =====
    # See: https://github.com/metadriverse/metadrive/issues/283
    success_reward=10.0,
    out_of_road_penalty=10.0,
    crash_vehicle_penalty=10.0,
    crash_object_penalty=1.0,
    driving_reward=1.0,
    speed_reward=0.1,
    use_lateral_reward=False,
    horizon=500,

    # ===== Cost Scheme =====
    crash_vehicle_cost=1.0,
    crash_object_cost=1.0,
    out_of_road_cost=1.0,

    # ===== Termination Scheme =====
    out_of_route_done=False,
    crash_vehicle_done=True,

    # ===== others =====
    interface_panel=["dashboard"],  # for boosting efficiency
)


class NuPlanEnv(BaseEnv):
    @classmethod
    def default_config(cls):
        config = super(NuPlanEnv, cls).default_config()
        config.update(NUPLAN_ENV_CONFIG)
        return config

    def __init__(self, config=None):
        super(NuPlanEnv, self).__init__(config)

    def _get_observations(self):
        return {self.DEFAULT_AGENT: self.get_single_observation()}

    def get_single_observation(self):
        if self.config["use_nuplan_observation"]:
            o = NuPlanObservation(self.config)
        else:
            o = LidarStateObservation(self.config)
        return o

    def switch_to_top_down_view(self):
        self.main_camera.stop_track()

    def switch_to_third_person_view(self):
        if self.main_camera is None:
            return
        self.main_camera.reset()
        if self.config["prefer_track_agent"] is not None and self.config["prefer_track_agent"] in self.vehicles.keys():
            new_v = self.vehicles[self.config["prefer_track_agent"]]
            current_track_vehicle = new_v
        else:
            if self.main_camera.is_bird_view_camera():
                current_track_vehicle = self.current_track_vehicle
            else:
                vehicles = list(self.engine.agents.values())
                if len(vehicles) <= 1:
                    return
                if self.current_track_vehicle in vehicles:
                    vehicles.remove(self.current_track_vehicle)
                new_v = get_np_random().choice(vehicles)
                current_track_vehicle = new_v
        self.main_camera.track(current_track_vehicle)
        return

    def setup_engine(self):
        super(NuPlanEnv, self).setup_engine()
        self.engine.register_manager("data_manager", NuPlanDataManager())
        self.engine.register_manager("map_manager", NuPlanMapManager())
        if not (self.config["no_traffic"] and self.config["no_pedestrian"]):
            self.engine.register_manager("traffic_manager", NuPlanTrafficManager())
        if not self.config["no_light"]:
            self.engine.register_manager("light_manager", NuPlanLightManager())
        self.engine.accept("q", self.switch_to_third_person_view)
        self.engine.accept("b", self.switch_to_top_down_view)
        self.engine.accept("]", self.next_seed_reset)
        self.engine.accept("[", self.last_seed_reset)

    def next_seed_reset(self):
        if self.current_seed + 1 < self.config["start_scenario_index"] + self.config["num_scenarios"]:
            self.reset(self.current_seed + 1)
        else:
            logging.warning("Can't load next scenario! current seed is already the max scenario index")

    def last_seed_reset(self):
        if self.current_seed - 1 >= self.config["start_scenario_index"]:
            self.reset(self.current_seed - 1)
        else:
            logging.warning("Can't load last scenario! current seed is already the min scenario index")

    def step(self, actions):
        ret = super(NuPlanEnv, self).step(actions)
        return ret

    def done_function(self, vehicle_id: str):
        vehicle = self.vehicles[vehicle_id]
        done = False
        done_info = dict(
            crash_vehicle=False, crash_object=False, crash_building=False, out_of_road=False, arrive_dest=False
        )
        # if np.linalg.norm(vehicle.position - self.engine.map_manager.sdc_dest_point) < 5 \
        #         or vehicle.lane.index in self.engine.map_manager.sdc_destinations:
        if self._is_arrive_destination(vehicle):
            done = True
            logging.info("Episode ended! Reason: arrive_dest.")
            done_info[TerminationState.SUCCESS] = True
        if self._is_out_of_road(vehicle):
            done = True
            logging.info("Episode ended! Reason: out_of_road.")
            done_info[TerminationState.OUT_OF_ROAD] = True
        if vehicle.crash_vehicle and self.config["crash_vehicle_done"]:
            done = True
            logging.info("Episode ended! Reason: crash vehicle ")
            done_info[TerminationState.CRASH_VEHICLE] = True
        if vehicle.crash_object:
            done = True
            done_info[TerminationState.CRASH_OBJECT] = True
            logging.info("Episode ended! Reason: crash object ")
        if vehicle.crash_building:
            done = True
            done_info[TerminationState.CRASH_BUILDING] = True
            logging.info("Episode ended! Reason: crash building ")

        # for compatibility
        # crash almost equals to crashing with vehicles
        done_info[TerminationState.CRASH] = (
            done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.CRASH_OBJECT]
            or done_info[TerminationState.CRASH_BUILDING]
        )
        return done, done_info

    def cost_function(self, vehicle_id: str):
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()
        step_info["cost"] = 0
        if self._is_out_of_road(vehicle):
            step_info["cost"] = self.config["out_of_road_cost"]
        elif vehicle.crash_vehicle:
            step_info["cost"] = self.config["crash_vehicle_cost"]
        elif vehicle.crash_object:
            step_info["cost"] = self.config["crash_object_cost"]
        return step_info['cost'], step_info

    def reward_function(self, vehicle_id: str):
        """
        Override this func to get a new reward function
        :param vehicle_id: id of BaseVehicle
        :return: reward
        """
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()

        # Reward for moving forward in current lane
        if vehicle.lane in vehicle.navigation.current_ref_lanes:
            current_lane = vehicle.lane
        else:
            current_lane = vehicle.navigation.current_ref_lanes[0]
        long_last, _ = current_lane.local_coordinates(vehicle.last_position)
        long_now, lateral_now = current_lane.local_coordinates(vehicle.position)

        # update obs
        self.observations[vehicle_id].lateral_dist = \
            self.engine.map_manager.current_sdc_route.local_coordinates(vehicle.position)[-1]

        # reward for lane keeping, without it vehicle can learn to overtake but fail to keep in lane
        if self.config["use_lateral_reward"]:
            lateral_factor = clip(1 - 2 * abs(lateral_now) / 6, 0.0, 1.0)
        else:
            lateral_factor = 1.0

        reward = 0
        reward += self.config["driving_reward"] * (long_now - long_last) * lateral_factor

        step_info["step_reward"] = reward

        if self._is_arrive_destination(vehicle):
            reward = +self.config["success_reward"]
        elif self._is_out_of_road(vehicle):
            reward = -self.config["out_of_road_penalty"]
        elif vehicle.crash_vehicle:
            reward = -self.config["crash_vehicle_penalty"]
        elif vehicle.crash_object:
            reward = -self.config["crash_object_penalty"]
        return reward, step_info

    def _is_arrive_destination(self, vehicle):
        return True if np.linalg.norm(vehicle.position - self.engine.map_manager.sdc_dest_point) < 5 else False

    def _reset_global_seed(self, force_seed=None):
        if force_seed is not None:
            current_seed = force_seed
        elif self.config["sequential_seed"]:
            current_seed = self.engine.global_seed
            if current_seed is None:
                current_seed = self.config["start_scenario_index"]
            else:
                current_seed += 1
            if current_seed >= self.config["start_scenario_index"] + int(self.config["num_scenarios"]):
                current_seed = self.config["start_scenario_index"]
        else:
            current_seed = get_np_random(None).randint(
                self.config["start_scenario_index"],
                self.config["start_scenario_index"] + int(self.config["num_scenarios"])
            )

        assert self.config["start_scenario_index"] <= current_seed < \
               self.config["start_scenario_index"] + self.config["num_scenarios"], "Force seed range Error!"
        self.seed(current_seed)

    def _is_out_of_road(self, vehicle):
        # A specified function to determine whether this vehicle should be done.
        done = vehicle.crash_sidewalk or vehicle.on_yellow_continuous_line or vehicle.on_white_continuous_line
        if self.config["out_of_route_done"]:
            done = done or self.observations[self.DEFAULT_AGENT].lateral_dist > 10
        return done
        # ret = vehicle.crash_sidewalk
        # return ret


if __name__ == "__main__":
    env = NuPlanEnv(
        {
            "use_render": True,
            "agent_policy": NuPlanReplayEgoCarPolicy,
            # "manual_control": True,
            "no_traffic": False,
            "no_pedestrian": False,
            "no_light": False,
            # "debug": True,
            "debug_static_world": False,
            "debug_physics_world": False,
            "load_city_map": True,
            # "global_light": False,
            "window_size": (1200, 800),
            # "multi_thread_render_mode": "Cull/Draw",
            "start_scenario_index": 0,
            # "pstats": True,
            "num_scenarios": 30,
            "show_coordinates": False,
            "horizon": 1000,
            # "show_fps": False,
            "vehicle_config": dict(
                lidar=dict(num_lasers=120, distance=50, num_others=0),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50),
                # show_lidar=True
                show_navi_mark=False,
                show_dest_mark=False,
                no_wheel_friction=True,
            ),
            "show_interface": False,
            "show_logo": False,
            "force_render_fps": 40,
            "show_fps": True,
            "DATASET_PARAMS": [
                # builder setting
                "scenario_builder=nuplan_mini",
                "scenario_builder.scenario_mapping.subsample_ratio_override=0.5",  # 10 hz

                # filter
                "scenario_filter=all_scenarios",  # simulate only one log
                "scenario_filter.remove_invalid_goals=true",
                "scenario_filter.shuffle=true",
                "scenario_filter.log_names=['2021.07.16.20.45.29_veh-35_01095_01486']",
                # "scenario_filter.scenario_types={}".format(all_scenario_types),
                # "scenario_filter.scenario_tokens=[]",
                # "scenario_filter.map_names=[]",
                # "scenario_filter.num_scenarios_per_type=1",
                # "scenario_filter.limit_total_scenarios=1000",
                # "scenario_filter.expand_scenarios=true",
                # "scenario_filter.limit_scenarios_per_type=10",  # use 10 scenarios per scenario type
                "scenario_filter.timestamp_threshold_s=20",  # minial scenario duration (s)
            ],
            "show_mouse": True,
        }
    )
    success = []
    env.reset()
    for seed in range(len(env.engine.data_manager._nuplan_scenarios)):
        env.reset()
        # env.reset(seed)
        for i in range(env.engine.data_manager.current_scenario_length * 10):
            o, r, tc, tm, info = env.step([0, 0])
            env.render(text={"seed": env.current_seed})
            if info["replay_done"]:
                break
    sys.exit()

# cull/draw camera
# draw set_state
# 2021.09.16.15.12.03_veh-42_01037_01434: 8/14



from metadrive.envs.metadrive_env import MetaDriveEnv

# setup_logger(True)

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "num_scenarios": 5,
            "traffic_density": 0.1,
            "traffic_mode": "respawn",
            "start_seed": 5,
            # "controller": "joystick",
            "manual_control": True,
            "use_render": True,
            "use_AI_protector": True,
            "map": 30
        }
    )

    o, _ = env.reset()
    # env.engine.force_fps.toggle()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        text = {"save": env.save_mode}
        env.render(text=text)
        # if d:
        #     # print("Reset")
        #     env.reset()
    env.close()



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

setup_logger(debug=True)

if __name__ == "__main__":
    import numpy as np

    env = MetaDriveEnv(
        {
            "num_scenarios": 4,
            "traffic_density": 0.0,
            "use_render": True,
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
                BaseMap.GENERATE_CONFIG: "SSSSSSSSSSSSS",
            },
            "manual_control": True
        }
    )
    acc = [0, 1]
    brake = [-1, -np.nan]
    env.reset()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(acc)
        # print(
        #     "new:{}, old:{}, diff:{}".format(
        #         env.vehicle.speed_km_h, env.vehicle.system.get_current_speed_km_hour(),
        #         env.vehicle.speed_km_h - env.vehicle.system.get_current_speed_km_hour()
        #     )
        # )
        env.render("Test: {}".format(i))
    env.close()



import gymnasium as gym

from metadrive.engine import get_logger
from metadrive.envs import MetaDriveEnv
from metadrive.envs import MultiAgentTollgateEnv, MultiAgentBottleneckEnv, MultiAgentRoundaboutEnv, \
    MultiAgentIntersectionEnv, MultiAgentParkingLotEnv, MultiAgentMetaDrive
from metadrive.envs import SafeMetaDriveEnv

metadrive_environment_dict = {
    "MetaDrive-validation-v0": {
        "start_seed": 0,
        "num_scenarios": 1000
    },
    "MetaDrive-10env-v0": {
        "start_seed": 1000,
        "num_scenarios": 10
    },
    "MetaDrive-100envs-v0": {
        "start_seed": 1000,
        "num_scenarios": 100
    },
    "MetaDrive-1000envs-v0": {
        "start_seed": 1000,
        "num_scenarios": 1000
    },
    # "MetaDrive-test-v0": {
    #     "start_seed": 0,
    #     "num_scenarios": 200
    # },
    # "MetaDrive-training0-v0": {
    #     "start_seed": 3000,
    #     "num_scenarios": 1000
    # },
    # "MetaDrive-training1-v0": {
    #     "start_seed": 5000,
    #     "num_scenarios": 1000
    # },
    # "MetaDrive-training2-v0": {
    #     "start_seed": 7000,
    #     "num_scenarios": 1000
    # },
}

safe_metadrive_environment_dict = {
    "SafeMetaDrive-validation-v0": {
        "start_seed": 0,
        "num_scenarios": 100
    },
    "SafeMetaDrive-10env-v0": {
        "start_seed": 1000,
        "num_scenarios": 10
    },
    "SafeMetaDrive-100envs-v0": {
        "start_seed": 1000,
        "num_scenarios": 100
    },
    "SafeMetaDrive-1000envs-v0": {
        "start_seed": 1000,
        "num_scenarios": 1000
    },
}

marl_env = {
    "MARLTollgate-v0": MultiAgentTollgateEnv,
    "MARLBottleneck-v0": MultiAgentBottleneckEnv,
    "MARLRoundabout-v0": MultiAgentRoundaboutEnv,
    "MARLIntersection-v0": MultiAgentIntersectionEnv,
    "MARLParkingLot-v0": MultiAgentParkingLotEnv,
    "MARLMetaDrive-v0": MultiAgentMetaDrive
}

envs = []
for env_name, env_config in metadrive_environment_dict.items():
    if env_name not in gym.registry.keys():
        envs.append(env_name)
        gym.register(id=env_name, entry_point=MetaDriveEnv, kwargs=dict(config=env_config))

for env_name, env_config in safe_metadrive_environment_dict.items():
    if env_name not in gym.registry.keys():
        envs.append(env_name)
        gym.register(id=env_name, entry_point=SafeMetaDriveEnv, kwargs=dict(config=env_config))

for env_name, entry in marl_env.items():
    if env_name not in gym.registry.keys():
        envs.append(env_name)
        gym.register(id=env_name, entry_point=entry, kwargs=dict(config={}))

if len(envs) > 0:
    get_logger("Gym").info("Successfully registered the following environments: {}.".format(envs))

if __name__ == '__main__':
    # Test purpose only
    env = gym.make("MetaDrive-validation-v0")
    env.reset()
    env.close()

    env = gym.make("SafeMetaDrive-validation-v0")
    env.reset()
    env.close()

    env = gym.make("MARLTollgate-v0")
    env.reset()
    env.close()



from abc import ABC
from copy import deepcopy


class ObservationBase(ABC):
    def __init__(self, config):
        # assert not engine_initialized(), "Observations can not be created after initializing the simulation"
        self.config = deepcopy(config)
        self.current_observation = None

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()

    @property
    def observation_space(self):
        raise NotImplementedError

    def observe(self, *args, **kwargs):
        raise NotImplementedError

    def reset(self, env, vehicle=None):
        pass



from panda3d.core import *
from direct.directbase import DirectStart
import random


def makeGrassBlades():
    format = GeomVertexFormat.getV3n3cpt2()
    vdata = GeomVertexData('blade', format, Geom.UHStatic)
    vertex = GeomVertexWriter(vdata, 'vertex')
    normal = GeomVertexWriter(vdata, 'normal')
    texcoord = GeomVertexWriter(vdata, 'texcoord')
    blade = Geom(vdata)

    for x in range(20):
        for y in range(20):
            r = random.uniform(0, 0.4)
            vertex.addData3f((x * 0.4) - 0.0291534 + r, (y * 0.4) + 0.0101984 + r, 0.0445018)
            vertex.addData3f((x * 0.4) + 0.0338934 + r, (y * 0.4) + 0.041644 + r, 0.83197)
            vertex.addData3f((x * 0.4) + 0.0304494 + r, (y * 0.4) - 0.00795362 + r, 0.360315)
            vertex.addData3f((x * 0.4) - 0.0432457 + r, (y * 0.4) - 0.0362444 + r, 0.0416673)
            vertex.addData3f((x * 0.4) - 0.0291534 + r, (y * 0.4) + 0.0101984 + r, 0.0445018)
            normal.addData3f(0.493197, 0.854242, -0.164399)
            normal.addData3f(-0.859338, 0.496139, -0.124035)
            normal.addData3f(-0.759642, -0.637797, -0.127114)
            normal.addData3f(0.974147, -0.0584713, -0.218218)
            normal.addData3f(0.493197, 0.854242, -0.164399)
            texcoord.addData2f(0.0478854, 0.000499576)
            texcoord.addData2f(0.353887, 0.9995)
            texcoord.addData2f(0.999501, 0.363477)
            texcoord.addData2f(0.729119, 0.000499576)
            texcoord.addData2f(0.000499547, 0.000499576)

    for z in range(0, 2000, 5):
        triangles = GeomTriangles(Geom.UHStatic)
        triangles.addVertices(0 + z, 1 + z, 2 + z)
        triangles.addVertices(2 + z, 3 + z, 0 + z)
        triangles.addVertices(1 + z, 4 + z, 2 + z)
        triangles.addVertices(3 + z, 2 + z, 4 + z)
        blade.addPrimitive(triangles)

    snode = GeomNode('node')
    snode.addGeom(blade)
    return snode


for x in range(4):
    for y in range(4):
        grass_group = render.attachNewNode(makeGrassBlades())
        grass_group.setTexture(loader.loadTexture("grass_mini.png"))
        grass_group.setColor(1, 1, 1, 1)
        grass_group.setPos(x * 8, y * 8, 0)

# grass_group.analyze()

# Create some lighting
ambientLight = AmbientLight("ambientLight")
ambientLight.setColor(Vec4(.3, .3, .3, 1))
directionalLight = DirectionalLight("directionalLight")
directionalLight.setDirection(Vec3(-5, -5, -5))
directionalLight.setColor(Vec4(1, 1, 1, 1))
directionalLight.setSpecularColor(Vec4(1, 1, 1, 1))
render.setLight(render.attachNewNode(directionalLight))
render.setLight(render.attachNewNode(ambientLight))
render.setShaderAuto()

run()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

from panda3d.core import SamplerState, Vec4

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.image import Image


class CubemapFilter(RPObject):
    """ Util class for filtering cubemaps, provides funcionality to generate
    a specular and diffuse IBL cubemap. """

    # Fixed size for the diffuse cubemap, since it does not contain much detail
    DIFFUSE_CUBEMAP_SIZE = 10
    PREFILTER_CUBEMAP_SIZE = 32

    def __init__(self, parent_stage, name="Cubemap", size=128):
        """ Inits the filter from a given stage """
        RPObject.__init__(self)
        self._stage = parent_stage
        self._name = name
        self._size = size
        self._make_maps()

    @property
    def specular_cubemap(self):
        """ Returns the generated specular cubemap. The specular cubemap is
        mipmapped and provides the specular IBL components of the input cubemap. """
        return self._specular_map

    @property
    def diffuse_cubemap(self):
        """ Returns the generated diffuse cubemap. The diffuse cubemap has no
        mipmaps and contains the filtered diffuse component of the input cubemap. """
        return self._diffuse_map

    @property
    def target_cubemap(self):
        """ Returns the target where the caller should write the initial cubemap
        data to """
        return self._specular_map

    @property
    def size(self):
        """ Returns the size of the created cubemap, previously passed to the
        constructor of the filter """
        return self._size

    def create(self):
        """ Creates the filter. The input cubemap should be mipmapped, and will
        get reused for the specular cubemap. """
        self._make_specular_targets()
        self._make_diffuse_target()

    def _make_maps(self):
        """ Internal method to create the cubemap storage """

        # Create the cubemaps for the diffuse and specular components
        self._prefilter_map = Image.create_cube(
            self._name + "IBLPrefDiff", CubemapFilter.PREFILTER_CUBEMAP_SIZE, "R11G11B10"
        )
        self._diffuse_map = Image.create_cube(self._name + "IBLDiff", CubemapFilter.DIFFUSE_CUBEMAP_SIZE, "R11G11B10")
        self._spec_pref_map = Image.create_cube(self._name + "IBLPrefSpec", self._size, "R11G11B10")
        self._specular_map = Image.create_cube(self._name + "IBLSpec", self._size, "R11G11B10")

        # Set the correct filtering modes
        for tex in [self._diffuse_map, self._specular_map, self._prefilter_map, self._spec_pref_map]:
            tex.set_minfilter(SamplerState.FT_linear)
            tex.set_magfilter(SamplerState.FT_linear)
            tex.set_clear_color(Vec4(0))
            tex.clear_image()

        # Use mipmaps for the specular cubemap
        self._spec_pref_map.set_minfilter(SamplerState.FT_linear_mipmap_linear)
        self._specular_map.set_minfilter(SamplerState.FT_linear_mipmap_linear)

    def _make_specular_targets(self):
        """ Internal method to create the specular mip chain """
        self._targets_spec = []
        self._targets_spec_filter = []
        mip, mipsize = 0, self._size
        while mipsize >= 2:

            # Assuming an initial size of a power of 2, its safe to do this
            mipsize = mipsize // 2

            # Create the target which downsamples the mipmap
            target = self._stage.create_target("CF:SpecIBL-PreFilter-" + str(mipsize))
            target.size = mipsize * 6, mipsize
            target.prepare_buffer()

            if mip == 0:
                target.set_shader_input("SourceTex", self._specular_map)
            else:
                target.set_shader_input("SourceTex", self._spec_pref_map)
            target.set_shader_input("DestMipmap", self._spec_pref_map, False, True, -1, mip + 1, 0)
            target.set_shader_input("currentMip", mip)

            mip += 1

            # Create the target which filters the mipmap and removes the noise
            target_filter = self._stage.create_target("CF:SpecIBL-PostFilter-" + str(mipsize))
            target_filter.size = mipsize * 6, mipsize
            target_filter.prepare_buffer()
            target_filter.set_shader_inputs(currentMip=mip, SourceTex=self._spec_pref_map)
            target_filter.set_shader_input("DestMipmap", self._specular_map, False, True, -1, mip, 0)

            self._targets_spec.append(target)
            self._targets_spec_filter.append(target_filter)

    def _make_diffuse_target(self):
        """ Internal method to create the diffuse cubemap """

        # Create the target which integrates the lambert brdf
        self._diffuse_target = self._stage.create_target("CF:DiffuseIBL")
        self._diffuse_target.size = (CubemapFilter.PREFILTER_CUBEMAP_SIZE * 6, CubemapFilter.PREFILTER_CUBEMAP_SIZE)
        self._diffuse_target.prepare_buffer()

        self._diffuse_target.set_shader_inputs(
            SourceCubemap=self._specular_map,
            DestCubemap=self._prefilter_map,
            cubeSize=CubemapFilter.PREFILTER_CUBEMAP_SIZE
        )

        # Create the target which removes the noise from the previous target,
        # which is introduced with importance sampling
        self._diff_filter_target = self._stage.create_target("CF:DiffPrefIBL")
        self._diff_filter_target.size = (CubemapFilter.DIFFUSE_CUBEMAP_SIZE * 6, CubemapFilter.DIFFUSE_CUBEMAP_SIZE)
        self._diff_filter_target.prepare_buffer()

        self._diff_filter_target.set_shader_inputs(
            SourceCubemap=self._prefilter_map,
            DestCubemap=self._diffuse_map,
            cubeSize=CubemapFilter.DIFFUSE_CUBEMAP_SIZE
        )

    def reload_shaders(self):
        """ Sets all required shaders on the filter. """

        # Set diffuse filter shaders
        self._diffuse_target.shader = self._stage.load_shader("ibl/cubemap_diffuse.frag.glsl")
        self._diff_filter_target.shader = self._stage.load_shader("ibl/cubemap_diffuse_filter.frag.glsl")

        # Set specular prefilter shaders
        mip_shader = self._stage.load_shader("ibl/cubemap_specular_prefilter.frag.glsl")
        for target in self._targets_spec:
            target.shader = mip_shader

        # Special shader for the first prefilter target
        self._targets_spec[0].shader = self._stage.load_shader("ibl/cubemap_specular_prefilter_first.frag.glsl")

        # Set specular filter sampling shaders
        mip_filter_shader = self._stage.load_shader("ibl/cubemap_specular_filter.frag.glsl")
        for target in self._targets_spec_filter:
            target.shader = mip_filter_shader

        # Special shader for the first filter target
        self._targets_spec_filter[0].shader = self._stage.load_shader("ibl/cubemap_specular_filter_first.frag.glsl")



from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import CreateAdverseRoad, CreateRoadFrom, ExtendStraightLane, \
    create_wave_lanes
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class Bottleneck(PGBlock):
    """
    This block is used to change thr lane num
    """
    ID = None
    SOCKET_NUM = 1
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.BOTTLENECK_PARAMETER)

    # property of bottleneck
    BOTTLENECK_LEN = None

    def get_intermediate_spawn_lanes(self):
        """
        Only spawn on straight lane
        """
        lanes = super(Bottleneck, self).get_intermediate_spawn_lanes()
        filtered_lanes = []
        for lane in lanes:
            if isinstance(lane[0], StraightLane):
                filtered_lanes.append(lane)
        return filtered_lanes


class Merge(Bottleneck):
    """
    -----\
          \
           -------------------
           -------------------
          /
    -----/
    InBottlecneck
    """
    ID = "y"

    def _try_plug_into_previous_block(self) -> bool:
        no_cross = True
        parameters = self.get_config()
        center_line_type = PGLineType.CONTINUOUS if parameters["solid_center_line"] else PGLineType.BROKEN
        self.BOTTLENECK_LEN = parameters["bottle_len"]
        lane_num_changed = parameters[Parameter.lane_num]

        start_node = self.pre_block_socket.positive_road.end_node
        straight_lane_num = int(self.positive_lane_num - lane_num_changed)
        straight_lane_num = max(1, straight_lane_num)

        circular_lane_num = self.positive_lane_num - straight_lane_num

        # part 1, straight road 0
        basic_lane = self.positive_lanes[straight_lane_num - 1]
        ref_lane = ExtendStraightLane(basic_lane, self.BOTTLENECK_LEN, [PGLineType.NONE, PGLineType.NONE])
        straight_road = Road(start_node, self.road_node(0, 0))
        no_cross = CreateRoadFrom(
            ref_lane,
            straight_lane_num,
            straight_road,
            self.block_network,
            self._global_network,
            center_line_type=center_line_type,
            side_lane_line_type=PGLineType.SIDE if circular_lane_num == 0 else PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            straight_road,
            self.block_network,
            self._global_network,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.SIDE if circular_lane_num == 0 else PGLineType.NONE,
            center_line_type=center_line_type,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        # extend for socket ,part 1 road 1
        ref_lane = ExtendStraightLane(ref_lane, parameters[Parameter.length], [PGLineType.NONE, PGLineType.NONE])
        socket_road = Road(self.road_node(0, 0), self.road_node(0, 1))
        no_cross = CreateRoadFrom(
            ref_lane,
            straight_lane_num,
            socket_road,
            self.block_network,
            self._global_network,
            center_line_type=center_line_type,
            side_lane_line_type=PGLineType.SIDE,
            inner_lane_line_type=PGLineType.BROKEN,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            socket_road,
            self.block_network,
            self._global_network,
            inner_lane_line_type=PGLineType.BROKEN,
            side_lane_line_type=PGLineType.SIDE,
            center_line_type=center_line_type,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        negative_socket_road = -socket_road
        self.add_sockets(PGBlockSocket(socket_road, negative_socket_road))

        # part 2, circular part
        for index, lane in enumerate(self.positive_lanes[straight_lane_num:], 1):
            lateral_dist = index * self.lane_width / 2
            inner_node = self.road_node(1, index)
            side_line_type = PGLineType.SIDE if index == self.positive_lane_num - straight_lane_num else PGLineType.NONE

            # positive part
            circular_1, circular_2, _ = create_wave_lanes(lane, lateral_dist, self.BOTTLENECK_LEN, 5, self.lane_width)
            road_1 = Road(start_node, inner_node)
            no_cross = CreateRoadFrom(
                circular_1,
                1,
                road_1,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross
            road_2 = Road(inner_node, self.road_node(0, 0))
            no_cross = CreateRoadFrom(
                circular_2,
                1,
                road_2,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross

            # adverse part
            lane = negative_socket_road.get_lanes(self.block_network)[-1]
            circular_2, circular_1, _ = create_wave_lanes(
                lane, lateral_dist, self.BOTTLENECK_LEN, 5, self.lane_width, False
            )
            road_2 = -road_2
            no_cross = CreateRoadFrom(
                circular_2,
                1,
                road_2,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross

            road_1 = -road_1
            no_cross = CreateRoadFrom(
                circular_1,
                1,
                road_1,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross

        return no_cross


class Split(Bottleneck):
    """
                        /-----
                       /
    -------------------
    -------------------
                       \
                        \-----
    OutBottlecneck
    """
    ID = "Y"

    def _try_plug_into_previous_block(self) -> bool:
        no_cross = True
        parameters = self.get_config()
        self.BOTTLENECK_LEN = parameters["bottle_len"]
        center_line_type = PGLineType.CONTINUOUS if parameters["solid_center_line"] else PGLineType.BROKEN
        lane_num_changed = parameters[Parameter.lane_num]

        start_node = self.pre_block_socket.positive_road.end_node
        straight_lane_num = self.positive_lane_num
        circular_lane_num = lane_num_changed
        total_num = straight_lane_num + circular_lane_num

        # part 1, straight road 0
        basic_lane = self.positive_lanes[straight_lane_num - 1]
        ref_lane = ExtendStraightLane(basic_lane, self.BOTTLENECK_LEN, [PGLineType.NONE, PGLineType.NONE])
        straight_road = Road(start_node, self.road_node(0, 0))
        no_cross = CreateRoadFrom(
            ref_lane,
            straight_lane_num,
            straight_road,
            self.block_network,
            self._global_network,
            center_line_type=center_line_type,
            side_lane_line_type=PGLineType.NONE,
            inner_lane_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            straight_road,
            self.block_network,
            self._global_network,
            inner_lane_line_type=PGLineType.NONE,
            side_lane_line_type=PGLineType.NONE,
            center_line_type=center_line_type,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        # part 2, circular part
        lane = self.positive_lanes[-1]
        socket_road_ref_lane = None
        for index in range(1, circular_lane_num + 1):
            lateral_dist = index * self.lane_width / 2
            inner_node = self.road_node(1, index)
            side_line_type = PGLineType.SIDE if index == circular_lane_num else PGLineType.NONE

            # positive part
            circular_1, circular_2, straight = create_wave_lanes(
                lane, lateral_dist, self.BOTTLENECK_LEN, parameters[Parameter.length], self.lane_width, False
            )
            if index == circular_lane_num:
                socket_road_ref_lane = straight
            road_1 = Road(start_node, inner_node)
            no_cross = CreateRoadFrom(
                circular_1,
                1,
                road_1,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross
            road_2 = Road(inner_node, self.road_node(0, 0))
            no_cross = CreateRoadFrom(
                circular_2,
                1,
                road_2,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross

        # extend for socket ,part 1 road 1
        socket_road = Road(self.road_node(0, 0), self.road_node(0, 1))
        no_cross = CreateRoadFrom(
            socket_road_ref_lane,
            total_num,
            socket_road,
            self.block_network,
            self._global_network,
            center_line_type=PGLineType.CONTINUOUS,
            side_lane_line_type=PGLineType.SIDE,
            inner_lane_line_type=PGLineType.BROKEN,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            socket_road,
            self.block_network,
            self._global_network,
            inner_lane_line_type=PGLineType.BROKEN,
            side_lane_line_type=PGLineType.SIDE,
            center_line_type=PGLineType.CONTINUOUS,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        negative_socket_road = -socket_road
        self.add_sockets(PGBlockSocket(socket_road, negative_socket_road))

        # part 2, circular part
        lanes = negative_socket_road.get_lanes(self.block_network)
        for index, lane in enumerate(lanes[self.positive_lane_num:], 1):
            lateral_dist = index * self.lane_width / 2
            inner_node = self.road_node(1, index)
            side_line_type = PGLineType.SIDE if index == circular_lane_num else PGLineType.NONE

            # positive part
            circular_1, circular_2, _ = create_wave_lanes(lane, lateral_dist, self.BOTTLENECK_LEN, 5, self.lane_width)
            road_1 = -Road(inner_node, self.road_node(0, 0))

            no_cross = CreateRoadFrom(
                circular_1,
                1,
                road_1,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross
            road_2 = -Road(start_node, inner_node)
            no_cross = CreateRoadFrom(
                circular_2,
                1,
                road_2,
                self.block_network,
                self._global_network,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=side_line_type,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            ) and no_cross
        return no_cross



#!/usr/bin/env python3
# Copyright (c) 2022 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse

from metadrive.third_party.kitsunetsuki.lut import Palette2LUT

from panda3d.core import Filename, VirtualFileSystem, get_model_path


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--output', type=str, help='output file')
    parser.add_argument('--input', type=str, help='input file', default=None)
    return parser.parse_args()


def main():
    args = parse_args()

    pl = Palette2LUT()
    pl.convert(args.input)
    pl.save(args.output)


if __name__ == '__main__':
    main()
